//package tlm;

/*
 * <p>Title: ClassTimer</p>
 * <p>Description: times duration of scheduled courses for TLM</p>
 * <p>Company: GMU NETLAB</p>
 * @author JMP
 * @version 4.4.2
 * @modified JMP 3-27-09
 */

/*----------------------------------------------------------------*
|   Copyright 2001-2009 Networking and Simulation Laboratory      |
|         George Mason University, Fairfax, Virginia              |
|                                                                 |
| Permission to use, copy, modify, and distribute this            |
| software and its documentation for academic purposes is hereby  |
| granted without fee, provided that the above copyright notice   |
| and this permission appear in all copies and in supporting      |
| documentation, and that the name of George Mason University     |
| not be used in advertising or publicity pertaining to           |
| distribution of the software without specific, written prior    |
| permission. GMU makes no representations about the suitability  |
| of this software for any purposes.  It is provided "AS IS"      |
| without express or implied warranties.  All risk associated     |
| with use of this software is expressly assumed by the user.     |
 *----------------------------------------------------------------*/

 // Operational note: our user documentation states that "If the
 // 'courselist' schedule is changed, the server must be restarted
 // for the change to take effect."   Given the code below, the
 // actual situation is less restrictive than this, but we judge it
 // would be too hard to communicate. Less simply stated, here is
 // how the actual code works:
 //
 // When TLM starts in autoSchedule mode, it finds the next course
 // in 'courseinfo' scheduled for the server, and locks into that
 // schedule. No lockin is created beyond that; instead the next
 // course will be found from the database at the end of the
 // currently scheduled course. Thus it is possible to change
 // the database for any course that is not already locked in as
 // next, and have that change take effect without restarting the
 // server.

import java.io.*;

public class CourseTimer extends TLM implements Runnable
{
  public CourseTimer()
  {
    // start verification courseID at configuration file value
    thisCourseID = roomID;
  }

  public void run()
  {
    // sleep until time for class (this has the effect that users
    // associated with the courseID = CID can login); then kick all users
    // off and accept only those associated with the scheduled course
    byte[] endTimeMessage = "scheduled class time over".getBytes();
    byte[] restartMessage = "server restarting for scheduled class".getBytes();

    // whenever the database doesn't hold any schedules,
    // there is nothing more for the CourseTimer to do
    while(Auth.getNextClassTime() > 0)
    {
      // shut down all connections
      classInSession = true;
      fcConnect.writeMessageToAll(34,restartMessage.length,restartMessage, false);
      closeAllConnections("closing current class");
      System.out.println("logged out users at " + dateTime());

      // load the generic room download buffer, if there is one
      thisCourseID = roomID;
      if(thisSemester.trim().length() > 0)
      {
        appFileDownloadName =
          new String("APPDATA" + thisSemester + thisCourseNum + ".zip");
  	    String appFilePathName =
  	      new String(appFilePath + appFileDownloadName);
System.out.println("*** APPFILE:"+appFilePathName);//debugx
  	    File appZipFile = new File(appFilePathName);
  	    if(appZipFile.exists())
          loadDownloadBuffer();
      }
	  
	  // setup directory for server-side recording, if any
	  if(recordOnServerConfig != 0)
	  {
		File serverRecordingsDirectory = new File(recordingsDirectory);
		if(!serverRecordingsDirectory.exists())
		{
		  System.out.println(
		    "error: recorderDataport configured but no serverRecordings directory found:" +
		    serverRecordingsDirectory);
		  System.out.println("server recording disabled for this course");
		  recordOnServer = false;
		  serverRecord = null;
		}
	  }

      // sleep until time for next class, allowing login to any user with account
      classInSession = false;
	  recordOnServer = false;
      System.out.println("server running with course/conference ID:" + roomID +
                         " at " + dateTime());
      System.out.println("next course/conference ID {" + printCourseID() +
                         "} starts on " + printCourseDayCode + " at " +
                         printStartTime + " ends at " + printEndTime);
      long sleepTime = nextCourseStart - System.currentTimeMillis();
      try
      {
        if(sleepTime > 0)Thread.sleep(sleepTime);
      }
      catch(InterruptedException ie){}

      // shut down all connections and change to scheduled class
      classInSession = true;
      fcConnect.writeMessageToAll(34,restartMessage.length,restartMessage, false);
      closeAllConnections("starting new class");
      System.out.println("logged out users and set course/conference ID to {" +
                         printCourseID() + "} at " + dateTime());
						 
      // shutdown the recorder
	  if(recordOnServerConfig != 0)
	  {
		if(serverRecord != null)
		  serverRecord.shutdownRecorder();
		serverRecord = null;
		recordOnServer = true;
	  }
      
      // load the download buffer, if there is one
      thisCourseID = nextCourseID[0];
      thisCourseNum = nextCourseNum[0];
      thisCourseNumBytes = thisCourseNum.getBytes();
      if(thisSemester.trim().length() > 0)
      {
        appFileDownloadName = new String("APPDATA" + 
		  thisSemester + thisCourseNum + ".zip");
        loadDownloadBuffer();
      }
	  
	  // start the recorder if configured
	  if(recordOnServer)
	  {
		serverRecord = new ServerRecord();
		
		// if startup failed, disable recorder
		if(!recordOnServer)serverRecord = null;
	  }

      // sleep until 10 minutes before shutdown
      sleepTime = nextCourseEnd - System.currentTimeMillis() - 600000;
      try
      {
        if(sleepTime > 0)Thread.sleep(sleepTime);
      }
      catch(InterruptedException ie){}
      sleepTime = nextCourseEnd - System.currentTimeMillis();
      long sleepMinutes = sleepTime/60000;

      // send a warning to a floor controllers, then sleep until end time
      byte[] warningMessage =
        ("server disconnect in " + sleepMinutes + " minutes").getBytes();
      fcConnect.writeMessageToAll(34,warningMessage.length,warningMessage,true);
      try
      {
        if(sleepTime > 0)Thread.sleep(sleepTime);
        else
        {
          System.out.println("autoSchedule panic break for negative sleepTime:" +
                             sleepTime);
          break;
        }
      }
      catch(InterruptedException ie){}
    } // end while(Auth.getNextClassTime() > 0)
    classInSession = false;
  } // end run()

  // method to build a String consisting of all values in nextCourseID
  static String printCourseID()
  {
    String printString = "";
    for(int printCount = 0; printCount < numberOfCourseIDAliases; ++printCount)
    {
      printString = printString.concat(nextCourseID[printCount]);
      printString = printString.concat(",");
    }
    // trim off the final comma
    if(printString.length() == 0)return "";
    return printString.substring(0,printString.length()-1);

  } // end printCourseID()

} // end class CourseTimer
