/* 
 * Heartbeat.java
 * Transport Layer Multicaster
 * Created on August 31, 2001
 * @author  mpullen & students
 * @version 5.2.4
 * @modified JMP 2/28/13
 */

/*----------------------------------------------------------------*
|   Copyright 2001-2013 Networking and Simulation Laboratory      |
|         George Mason University, Fairfax, Virginia              |
|                                                                 |
| Permission to use, copy, modify, and distribute this            |
| software and its documentation for all purposes is hereby       |
| granted without fee, provided that the above copyright notice   |
| and this permission appear in all copies and in supporting      |
| documentation, and that the name of George Mason University     |
| not be used in advertising or publicity pertaining to           |
| distribution of the software without specific, written prior    |
| permission. GMU makes no representations about the suitability  |
| of this software for any purposes.  It is provided "AS IS"      |
| without express or implied warranties.  All risk associated     |
| with use of this software is expressly assumed by the user.     |
*----------------------------------------------------------------*/

// This heartbeat is needed to wake up exception processing
// when a host disconnects.

/*********************SYSTEM LOGIC**************************
 *
 * Heartbeat.java class logic:
 * ->In an infinite loop
 * ->  generate a UDP message to loopback address/uPort[0]
 * ->  sleep one second
 ***********************************************************/

//package TLM;
import java.io.*;
import java.net.*;
import java.util.*;

public class Heartbeat extends TLM implements Runnable
{
  final static int heartbeatPeriod = 1000;
  public void run()
  {
	  // create a dummy buffer so we have something to send
	  byte[] heartbeatBuffer = {68, 0, 0, 0, 0};
    byte[] fcHeartbeat = {50, 0, 0};
    long lastHeartbeatTime = 0;
    long allowableHeartbeatInterval = 1500;

    // send a message once per second
    int hostIndex = 0;
    while (true)
    {
      try
      {
        // time to compare
        long timeNow = System.currentTimeMillis();
        long deltaTime = timeNow - lastHeartbeatTime;
        if((lastHeartbeatTime > 0) && (deltaTime > allowableHeartbeatInterval))
          System.out.println("*** " + dateTime() +
                             " unacceptable time between heartbeats:"  + deltaTime);
        
        // test all active hosts
        for (hostIndex = 0; hostIndex < numberOfHosts; hostIndex++)
        {
	        //debugx if (authenticatedHost[hostIndex] || connectedByTCP[hostIndex])
          if (tcpIsOpen[hostIndex] || udpActive[hostIndex])
		      {
			      // no TLMC RTT probe or other message received-
			      // assume connection or client is dead
			      if ( (timeNow - uRecvTime[hostIndex]) > tlmcDisconnectTime)
			      
			        // exception: with multicast FC there is nothing to start TLMC heartbeat
			        if(!fcIsMc[hostIndex])
			        {
			          closeTunnel(hostIndex, true,
						                "Heartbeat found expired TLMC receive time");
                continue;
			        }
              
            // FC heartbeat is tested only if there is a direct connection
			      if(fRecvTime[hostIndex] >= 0)
			      {
			        // no FC heartbeat received- assume connection or client is dead
			        if ( (timeNow - fRecvTime[hostIndex]) > fcDisconnectTime)
			        {
				        closeTunnel(hostIndex, true,
							              "Heartbeat found expired FC receive time");
				        continue;
			        }
                
			        // send a heartbeat to FC - if not possible, connection is broken
			        if ( (fcOutStreams[hostIndex] != null))
              {
				        try
				        {
				          fcOutStreams[hostIndex].write(fcHeartbeat, 0, 3);
				          fcOutStreams[hostIndex].flush();
                }
				        catch (Exception e)
				        {
				          closeTunnel(hostIndex, true, "Heartbeat found FC disconnect");
				          continue;
			          }
			        }
			        else
			        {
				        // if we have multicast floor control this isn't a problem
				        if(!usingMcFC)
				        closeTunnel(hostIndex, true, "Heartbeat found no FC connection");
                continue;
			        }
            }// end if(fRecvTime[hostIndex] >= 0)
          }// end if(authenticatedHost[ hostindex ] )
            
		      // if this host has a TCP tunnel open, send it a heartbeat
          queueTunnel(2,hostIndex,heartbeatBuffer,5);
          //debugxx if(writeOutStream(hostIndex,heartbeatBuffer,5))
          //debugxx   closeTunnel(hostIndex, true, "heartbeat tcpOutStreams disconnect");
          
        }// end for( int sendIndex...
    
        // wait one second and send again
        Thread.sleep(heartbeatPeriod);
      }  
      catch (Exception e)
      {
        closeTunnel(hostIndex,true, "exception in Heartbeat:" + e);
        e.printStackTrace(System.out);
      }
    } // end got(hostIndex...
  } // end run()
} // end class Heartbeat
