/*
 * MCFCListen.java
 * Transport Layer Multicaster
 * Created on February 16, 2007
 * @author mpullen
 * @version 4.4.2
 * @modified 12/28/07 by JMP
 */
/*----------------------------------------------------------------*
|   Copyright 2001-2008 Networking and Simulation Laboratory      |
|         George Mason University, Fairfax, Virginia              |
|                                                                 |
| Permission to use, copy, modify, and distribute this            |
| software and its documentation for academic purposes is hereby  |
| granted without fee, provided that the above copyright notice   |
| and this permission appear in all copies and in supporting      |
| documentation, and that the name of George Mason University     |
| not be used in advertising or publicity pertaining to           |
| distribution of the software without specific, written prior    |
| permission. GMU makes no representations about the suitability  |
| of this software for any purposes.  It is provided "AS IS"      |
| without express or implied warranties.  All risk associated     |
| with use of this software is expressly assumed by the user.     |
 *----------------------------------------------------------------*/



/***********************SYSTEM LOGIC*************************************
 * Instantiate multicast socket for give address/port
 * MCFCListen.java class logic:
 * ->In an infinite loop
 * ->  read multicast UDP messages from FC group & port
 * ->  has message source IP address
 * ->  if not in hashtable, message mush be Session Variable; else reject
 * ->  if Session Variable, add IP address to the hashtable
 *     forward message to Floor Control methods in FCConnect
 *
************************************************************************/
//package TLM;
import java.io.*;
import java.net.*;
import java.util.*;


public class MCFCListen extends TLM //implements Runnable
{
  static InetAddress localAddress;

  // sourceAddress is the Inetaddress from which the message is received
  // localAddress is the Inetaddress of the host on which TLM is running

  // validated hosts contains IP addresses of all hosts from which we have
  // received multicast floor control messages with valid session variable
  static Hashtable validatedHosts;
  byte[] recvBuffer;

  // constructor
  public MCFCListen()
  {
/* irmp
    try
    {
      // open sockets
      mcFcSocket = new MulticastSocket(mcFcPort);
      mcFcSocket.joinGroup(mcFcInetAddr);
    }
    catch(IOException ioe)
    {
      System.out.println(
        "***** IOException opening/joining multicast floor control:" +
        ioe.getMessage());
      done = true;
    }
    try
    {
      // get address of the local host
      localAddress = InetAddress.getLocalHost();
    }
    catch(UnknownHostException e)
    {
      System.out.println("****** Unknown local host error"+e);
      done = true;
    }
*/
    //  initialize and allocate buffers
    recvBuffer = new byte[ MAX_MESSAGE ]; // receives floor control headers

    // initialize the hashtable
    validatedHosts = new Hashtable(2*maxConnections);
  }

  int hostIndex = -1, messageLength = 0;
  byte[] rejectMessage = {28,0,0};
/* irmp
  public void run()
  {
    // perform this loop as long as floor control messages arrive
    try
    {
      while (!done)
      {
        // give up control so another thread can run
        try
        {
          Thread.sleep(sleepTime);
        }
        catch (InterruptedException e) {}

System.out.println("ready to receive mc on "+mcFcInetAddr
+"/"+mcFcSocket.getLocalPort());//debugx
        // receive a multicast message
        DatagramPacket fcMessage = new DatagramPacket(recvBuffer, MAX_MESSAGE);
        mcFcSocket.receive(fcMessage);
        MCFCReceive(recvBuffer,
                    fcMessage.getLength(),
                    fcMessage.getAddress());
      }// end while (!done)
    }
    catch(IOException ioe)
    {
      System.out.println("IOException reading multicast floor control input");
      return;
    }

  } // end run
irmp */


  // process an incoming multicast FC message, which could come
  // either direct from MCFCListen run() or from IRMP
  void MCFCReceive(byte[] recvBuffer,
                   int messageLength,
                   InetAddress sourceAddress)
  {
     // convert host address to string format
     String sourceHostAddress = sourceAddress.getHostAddress();

     // if not using Web authentication we will need to force authentication
     boolean forceAuthentication = false;

     // don't respond to loopback
     //if(sourceHostAddress.equals(localHostAddress))return;/debugx

     // keep track of number of messages received
     countMcRecv++;

     // validate length of received message
     if (messageLength < 3)
     {
       System.out.println(
         "received multicast floor control message of invalid length:" +
         messageLength);
       return;
     }
     // extract FC message payload length
     int fcLength = bytesToUshort(recvBuffer[1], recvBuffer[2]);
     if (fcLength > 255)
     {
       System.out.println(
         "received multicast floor control message with invalid imbedded length:" +
         fcLength);
       return;
     }
     // attempt to extract hostIndex from hashtable
     int messageCode = recvBuffer[0];
     Integer Index;
     if ( (Index = (Integer)validatedHosts.get(sourceHostAddress)) == null)
     {
       // ignore all but session variable messages
       if (recvBuffer[0] != 11)return;
       if (messageLength != 21)
       {
         System.out.println(
           "ill formed session variable message received from host " +
           sourceHostAddress + " message length:" + messageLength);
         return;
       }
       // check database for session variable; if no match, force authentication
       byte[] sessionVariableBytes = new byte[16];
       String testSessionVariable = new String(sessionVariableBytes);
       System.arraycopy(recvBuffer, 3, sessionVariableBytes, 0, 16);
       if (!auth.validateSessionVariable(testSessionVariable,
                                         sourceHostAddress))
         forceAuthentication = true;

       // look up the sourceHostAddress in hostAddr[]; if not present add it
       for (hostIndex = 0; hostIndex < maxConnections; ++hostIndex)
         if (sourceHostAddress.equals(hostAddr[hostIndex]))
           break;
       if (hostIndex >= maxConnections)// => sourceHostAddress not in host table
       {
         // allow only one insert to host table at a time
         synchronized(validatedHosts)
         {
           // add to host table
           hostIndex = addToHostTable(sourceAddress,
                                      testSessionVariable,
                                      -1,
                                      -1);
         }
         // deal with insertion failures
         if(hostIndex < 0)
         {
           System.out.println("attempted new floor control participant " +
                              sourceHostAddress +
                              " found full host table - rejected");
           ucFcSend(rejectMessage,sourceAddress);
           return;
         }
       }// end if (hostIndex >= maxConnections)
       if(forceAuthentication)authenticatedHost[hostIndex] = false;

       // add address to hashtable
       synchronized(validatedHosts)
       {
         validatedHosts.put(sourceHostAddress, new Integer(hostIndex));
       }
       // send the message to FCConnect for processing
       int tlmcTcpPort = bytesToUshort(recvBuffer[19],recvBuffer[20]);
       fcConnect.processSessionVariable(hostIndex,
                                        sessionVariableBytes,
                                        sourceHostAddress,
                                        tlmcTcpPort);
       return;
     } // end if ( (Index =...

     // sender is in host table at Index and has sent valid session message
     // next message from this sender should be login message
     hostIndex = Index.intValue();

     // ignore heartbeats and our own multicast-sent messages
     if(messageCode > 31)return;
     if(messageCode == 19)// heartbeat
     {
       fRecvTime[hostIndex] = System.currentTimeMillis();
       return;
     }
     // ignore multiple session variable messages
     if(messageCode == 11)return;
     if (recvBuffer[0] == 0)// FC login message
     {
       // ignore multiple login messages
       if(authenticatedHost[hostIndex])return;
       if(fcLength != 30)
       {
         System.out.println(
           "ill formed multicast floor control login message received from " +
           sourceHostAddress + " messageLength:" + messageLength);
         return;
       }
       byte[] loginMessage = new byte[30];
       System.arraycopy(recvBuffer,3,loginMessage,0,30);
       fcConnect.processLoginMessage(hostIndex,
                                     loginMessage,
                                     sessionVariable[hostIndex].getBytes());

       // if authentication succeeded, flag this host as MC FC
       if(authenticatedHost[hostIndex])fcIsMc[hostIndex] = true;
       return;
     }
     // if this is not a login message, confirm login has occurred
     if(!authenticatedHost[hostIndex])
     {
       System.out.println("multicast floor control message received from host " +
           sourceHostAddress + " before validating session variable message");
       ucFcSend(rejectMessage,sourceAddress);
       closeTunnel(hostIndex,true,"floor control traffic before login message");
       return;
     }
     // regular floor control message - send to FCConnect for processing
     byte[] newFcMessage = new byte[fcLength];
     System.arraycopy(recvBuffer,3,newFcMessage,0,fcLength);
     fcConnect.processFCMessage((int)recvBuffer[0],
                                newFcMessage,
                                fcLength,
                                hostIndex);
  }// end MCFCReceive()

  // method to send a message to the floor control multicast group
  // assumes multicast is open and ready to send
  // returns number of bytes sent, or -1 for failure
  int mcFcSend(byte[] message)
  {
    // insert zeros in message after the header
    // to take place of IP address in unicast
    byte[] newMessage = new byte[message.length+4];
    System.arraycopy(message,0,newMessage,0,3);
    for(int i=3; i<7; ++i)newMessage[i] = (byte)0;
    System.arraycopy(message,3,newMessage,7,message.length-3);
///* irmp
    // send via IRMP
    irmp.irmp_send(new String(newMessage));
//irmp */
/* irmp
    // send the message
    DatagramPacket sendPacket = new DatagramPacket(newMessage,
                                                   newMessage.length,
                                                   mcFcInetAddr,
                                                   mcFcPort);
    try
    {
      mcFcSocket.send(sendPacket);
    }
    catch(IOException ioe)
    {
      return -1;
    }
irmp */
System.out.println("MCFCsend TC:"+newMessage[0]+" LEN:"+message.length);//debugx
    return message.length;
  }

  // method to send a unicast floor control message to a particular host
  // assumes mcFcSocket is open and ready to send
  // these are unicast messages with transactionCode {0,1,3,4,18}
  // returns number of bytes sent, or -1 for failure
  int ucFcSend(byte[] message,InetAddress sendToAddress)
  {
    // insert IP address in message after the header
    // so target floor control can recognize it
    byte[] newMessage = new byte[message.length+4];
    System.arraycopy(message,0,newMessage,0,3);
    System.arraycopy(sendToAddress.getAddress(),0,newMessage,3,4);
    System.arraycopy(message,3,newMessage,7,message.length-3);
///* irmp
    // send via IRMP
    irmp.irmp_send(new String(newMessage));
System.out.println("UCFCsend TC:"+newMessage[0]+" LEN:"+message.length);//debugx
    return message.length;
//irmp */
/* irmp
    // make and send a datagram message
    DatagramPacket sendPacket = new DatagramPacket(newMessage,
                                                   newMessage.length,
                                                   mcFcInetAddr,
                                                   mcFcPort);
    try
    {
      mcFcSocket.send(sendPacket);
    }
    catch(IOException ioe)
    {
      return -1;
    }
    return newMessage.length;
irmp */
  }

  // method to remove a host from hashtable
  void removeHost(int hostIndex)
  {
    synchronized(validatedHosts)
    {
      if(hostAddr[hostIndex] != null)
        validatedHosts.remove(hostAddr[hostIndex]);
    }
  }

}// end MCFCListen class
