/* modified JMP 3/1/13
 * MCListen.java
 * Transport Layer Multicaster
 * Created on August 31, 2001
 * @author mpullen & students
 * @version 5.2.4
 */
/*----------------------------------------------------------------*
|   Copyright 2001-2013 Networking and Simulation Laboratory      |
|         George Mason University, Fairfax, Virginia              |
|                                                                 |
| Permission to use, copy, modify, and distribute this            |
| software and its documentation for academic purposes is hereby  |
| granted without fee, provided that the above copyright notice   |
| and this permission appear in all copies and in supporting      |
| documentation, and that the name of George Mason University     |
| not be used in advertising or publicity pertaining to           |
| distribution of the software without specific, written prior    |
| permission. GMU makes no representations about the suitability  |
| of this software for any purposes.  It is provided "AS IS"      |
| without express or implied warranties.  All risk associated     |
| with use of this software is expressly assumed by the user.     |
 *----------------------------------------------------------------*/



/***********************SYSTEM LOGIC****************** *
 * MCListen.java class logic:
 * ->In an infinite loop
 * ->  read multicast UDP messages
 * ->  update multicast last time recv
 * ->  if multicast UDP message is not from the bridge machine
 * ->  forward only if source is in unicast array
 *     (insert in array if a wildcard slot is available)
 * ->  also, forward only if floorHolder or port is not floor controlled
 * ->  send A copy to each unicast UDP address that has been active within time
 *       window unless port is associated with TCP on this host
 * ->  and to each unicast TCP address/TCP tunnel port
 *       (after prepending unicast UDP port to message)

*******************************************************/
//package TLM;
import java.io.*;
import java.net.*;

public class MCListen extends TLM implements Runnable
{
    MulticastSocket mcSocket;
    DatagramSocket udpSocket;
    DatagramPacket sendPacket;
    InetAddress sourceAddress;
    static InetAddress localAddress;
    // sourceAddress is Inetaddress from which the message is received
    // localAddress is the Inetaddress of the host on which TLM is running
    byte[] sendBuffer, recvBuffer;
    double window;           // window of time for unicast forwarding
    double[] mRecvTime;      // time when packet was last received from mAddr (only
                             // (position 0 is used- so as array this can be shared)
    int udpPort, mcPort;     // port numbers served
    int thisApplicationIndex = 0; // 0 = audio, 1 = WBD, 2 = video
    boolean nonVideoPort, forwardingAudio, forwardingWBD, forwardingVideo;
    private long startTime;// system time when any message is received
    private long checkTime = 0;

    // constructor
    public MCListen( int applicationIndex, int udpPortParm, int mcPortParm,
      MulticastSocket mcSocketParm, DatagramSocket udpSocketParm )
    {
      // copy input parameters and set related values
      udpPort = udpPortParm;
      mcPort = mcPortParm;
      mcSocket = mcSocketParm;
      udpSocket = udpSocketParm;
      nonVideoPort = ( (udpPort != videoPort) && (udpPort != videoPort+1) );
      forwardingAudio = ( (udpPort == audioPort) || (udpPort == audioPort+1) );
      if( forwardingAudio )thisApplicationIndex = 0;
      forwardingWBD = ( (udpPort == wbdPort) || (udpPort == wbdPort+1) );
      if( forwardingWBD )thisApplicationIndex = 1;
      forwardingVideo = ( (udpPort == videoPort) || (udpPort == videoPort+1) );
      if( forwardingVideo )thisApplicationIndex = 2;

      try
      {
        // get address of the local host
        localAddress =InetAddress.getLocalHost();
      }
      catch(UnknownHostException e)
      {
        System.out.println("Unknown local host error" + e);
      }

      //  initialize and allocate buffers
      recvBuffer = new byte[ MAX_MESSAGE ];
      sendPacket = new DatagramPacket( recvBuffer, 0, 0, localAddress, 0);
      mRecvTime = new double[ numberOfApplications ];
    }

    public void run()
    {
      int thisHostIndex, messageLength = 0;

      // perform this loop as long as messages arrive
      try
      {
        while( !stopTLM )
        {
          // give up control so another thread can run
          try
          {
            Thread.sleep(sleepTime);
          }
          catch( InterruptedException e)
          {
            System.out.println( "Interrupted Exception :" +e.getMessage() );
          }
          // receive a multicast message
          DatagramPacket recvMessage = new DatagramPacket( recvBuffer, MAX_MESSAGE );
          mcSocket.receive( recvMessage );
          countMcRecv++;
          messageLength = recvMessage.getLength();
          sourceAddress = recvMessage.getAddress();
          startTime = System.currentTimeMillis();
          mRecvTime[ thisApplicationIndex ] = startTime;

          // allocate send buffer to be queued tor TunnelSend
          sendBuffer = new byte[ messageLength+5 ];
/* debugx temporary for CS758
          // identify sender in hosts[] array
          for( thisHostIndex = 0; thisHostIndex < numberOfHosts; ++thisHostIndex )
            if( hostAddr[ thisHostIndex ] != null)
              if( sourceAddress.equals( hostAddr[ thisHostIndex ] ) )break;
          if( thisHostIndex >= numberOfHosts )continue; // while(!stopTLM)
           
          // floor control for WBD and video
          boolean okToForward = (thisHostIndex == floorHolder);
          if( forwardingWBD )
            // allow WBD repair requests even under floor control
            if( !wbdFloorControl || (recvBuffer[15]==2) )okToForward = true;
          if( forwardingVideo )
            if( !videoFloorControl )okToForward = true;
          if( forwardingAudio )
          {
            if( (floorHolder >= 0) && (thisHostIndex != floorHolder) )
              checkTime = startTime;
            else
            {
              if( (startTime - checkTime) > 5000 )
              {
                if( lastWarningTime[thisHostIndex] < (startTime - 20000) )
                {
                  if( sendFloorWarning(thisHostIndex, udpPort) )
                    return;
                  lastWarningTime[thisHostIndex] = startTime;
                } // end if( lastWarningTime
              } // end if( startTime
            } // end if( (floorHolder >= 0)...
          } // end if( forwardingAudio)
          
          // if not OK to forward, go back to read another
          if( !okToForward )continue;
debugx CS758 */
          // build outgoing TCP message; start message with protocol transaction code
          sendBuffer[0] = 66;
           
          // insert UDP message length
          // inline version of intToByte()
          // sendBuffer[1] = intToByte1( messageLength );
          // sendBuffer[2] = intToByte2( messageLength );
          // coded inline for efficiency
          int highByte = messageLength >> 8;
          if( highByte <= 127 )
            sendBuffer[1] = (byte) highByte;
          else
            sendBuffer[1] = (byte)(int)(127-highByte);
          int lowByte = messageLength - 256*highByte;
          if( lowByte <= 127 )
            sendBuffer[2] = (byte)lowByte;
          else
           sendBuffer[2] = (byte)(int)(127-lowByte);

          // insert UDP port number
          // sendBuffer[3] = intToByte1( udpPort );
          // sendBuffer[4] = intToByte2( udpPort );
          // coded inline for efficiency
          highByte = udpPort >> 8;
          if( highByte <= 127 )
            sendBuffer[3] = (byte) highByte;
          else
            sendBuffer[3] = (byte)(int)(127-highByte);
          lowByte = udpPort - 256*highByte;
          if( lowByte <= 127 )
            sendBuffer[4] = (byte)lowByte;
          else
            sendBuffer[4] = (byte)(int)(127-lowByte);

          // transfer remainder of message
          System.arraycopy(recvBuffer,0,sendBuffer,5,messageLength);

          // send a copy of the received TCP message to every host with open tunnel
          int tcpLength = messageLength + 5;
          for( int hostIndex = 0; hostIndex < numberOfHosts; ++hostIndex )
          {
            if( !connectedByTCP[ hostIndex ] )continue;

            // place message in queue for tunnel send
            if( forwardingAudio )
            {
              if( queueTunnel( 1, hostIndex, sendBuffer, tcpLength ) )
              {
                if( ++audioDropped[ hostIndex ]%dropMessageInterval == 1 )
                  System.out.println( dateTime() + " " +
                    audioDropped[ hostIndex ] + " audio messages for host #" +
                    hostIndex + " dropped due to congestion (delta RTT:" +
                    latestRTTIncrease[ hostIndex ] + ")" );
              }
              else if( forwardingWBD )
              {
                if( queueTunnel( 3, hostIndex, sendBuffer, tcpLength ) )
                {
                  if( ++whiteboardDropped[ hostIndex ]%dropMessageInterval == 1 )
                    System.out.println( dateTime() + " " +
                      whiteboardDropped[ hostIndex ] +
                      " whiteboard messages for host #" + hostIndex +
                      " dropped due to congestion (delta RTT:" +
                      latestRTTIncrease[ hostIndex ] + ")" );
                }
              }
            }
            else if(connectedVideoByTCP[ hostIndex ])
              if( forwardingVideo )
                if( queueTunnel( 4, hostIndex, sendBuffer, tcpLength ) )
                {
                  if( ++videoDropped[ hostIndex ]%dropMessageInterval == 1 )
                    System.out.println( dateTime() + " " +
                      videoDropped[ hostIndex ] + " video messages for host #" +
                      hostIndex + " dropped due to congestion (delta RTT:" +
                      latestRTTIncrease[ hostIndex ] + ")" );
                }
        } // end for(sendIndex...
        // **************** end tunnel queueing *******************************

        if( retransmitUDP )
        {
          sendPacket.setData(recvBuffer, 0, messageLength);
          sendPacket.setPort( udpPort );
          for( int hostIndex = 0; hostIndex < udpFlowsActive; hostIndex++ )
          {
            if( !udpActive[ hostIndex ] )continue;

            // don't send to a host that is closing
            if( tunnelClosing[ hostIndex ] )continue;

            // for UDP, forward only if the messages have arrived within window
            if(( startTime - uRecvTime[ hostIndex ]) < window)
            {
              // do not send message back to source or from the multicaster host
              if( hostAddr[ hostIndex ] != null )
                if( ( sourceAddress.equals( hostAddr[ hostIndex ] ) )
                  || ( sourceAddress.equals( localAddress )) )continue;
              // send UDP message
              sendPacket.setAddress(
                InetAddress.getByAddress(hostAddr[ hostIndex ].getBytes()) );
              udpSocket.send( sendPacket );
              countUdpSend++;
            } // end if(( startTime...
            else
            {
              // address has stopped sending; delete from active set
              deactivateFlow( hostIndex );
            } // end else/if((startTime...
          } // end for( sendIndex
        } // end if( retransUDP )
        Thread.yield();
      }// end while (!stopTLM
    } // end main try
    catch(IOException ioe)
    {
      // lost connection - can't continue
    }
    catch(Exception e)
    {
      System.out.println("Exception in MCListen:" + e + " "+e.getMessage());
      e.printStackTrace(System.out);
    }
    if(mcSocket != null)mcSocket.close();
  } // end run

  public static void close()
  {
    // no more packets to be sent
    stopTLM = true;
  }

}// end TLM.MCListen class
