/* 
 * ServerRecord.java
 * Transport Layer Multicaster
 * Created on 03/25/09
 * @author  mpullen & students
 * @version 5.2.4
 * @modified JMP 5/7/2011 for user download of server recording
 * @modified JMP 1/14/13 for posting of server recording
 */

/*----------------------------------------------------------------*
|  Copyright 2011-2013 Networking and Simulation Laboratory       |
|         George Mason University, Fairfax, Virginia              |
|                                                                 |
| Permission to use, copy, modify, and distribute this            |
| software and its documentation for academic purposes is hereby  |
| granted without fee, provided that the above copyright notice   |
| and this permission appear in all copies and in supporting      |
| documentation, and that the name of George Mason University     |
| not be used in advertising or publicity pertaining to           |
| distribution of the software without specific, written prior    |
| permission. GMU makes no representations about the suitability  |
| of this software for any purposes.  It is provided "AS IS"      |
| without express or implied warranties.  All risk associated     |
| with use of this software is expressly assumed by the user.     |
*----------------------------------------------------------------*/
//package TLM;
import java.sql.*;
import java.io.*;
import java.net.*;
import java.util.*;
// thread to make one recorder connection, timestamp and feed it data
// recorder is assumed to be running on the same host as TLM
// NEEDS DONE: add appDataDownloadFile into recording;
// capture floor control in TCPTunnel and UDPTunnel
public class ServerRecord extends TLM
{
  Socket dataSocket, controlSocket;
  BufferedOutputStream dataStream, controlStream;
  RecorderThread recorderThread;
  Thread runRecorder, runFileCopy;
  boolean terminateRecorder = false;
  long recordingStartTime = 0;
  RecordQueueElement recordQueue = null;
  
  public ServerRecord()
  {
    // capture start time
	  recordingStartTime= System.currentTimeMillis();
    
	  // start thread to write the messages to file
	  recorderThread = new RecorderThread();
	  runRecorder = new Thread(recorderThread);
	  
	  // run this thread with priority lower than the 
	  // rest of TLM, since once the timestamp is made
	  // it doesn't matter if there is a delay writing
	  // the message to file
	  runRecorder.setPriority(runRecorder.getPriority()-1);
	  runRecorder.start();
		
  }// end ServerRecord constructor
  
  // stop recording
  void shutdownRecorder()
  {
    terminateRecorder = true;
	  synchronized(runRecorder)
	  {
	    runRecorder.interrupt();
    }
    System.out.println("stopped recorder at " + dateTime());
    
    // check whether we might need to copy the file for Moodle
    if(recordingsDownloadDirectory == null)return;
    
    // check whether a recording for today exists in the
    // recordingsSubdirectory directory - assume it is from
    // the client; if none, copy the serverside recording
    // we just made to downloadRecordings
    
    // find today's date
    GregorianCalendar gcalendar = new GregorianCalendar();
    int year = gcalendar.get(Calendar.YEAR)%100;
    int month = gcalendar.get(Calendar.MONTH) + 1 ;
    int date = gcalendar.get(Calendar.DATE);
    String stryear = "" + ( year < 10 ? "0" + year : "" + year);
    String strmonth = "" + ( month < 10 ? "0" + month : "" + month);
    String strdate = "" + ( date < 10 ? "0" + date : "" + date);
    String todaysDate = new String(stryear + "." + strmonth + "." + strdate);
    
    // get the names of files in downloadDirectory
    File dlDirectory = new File(recordingsDownloadDirectory);
    File[] checkFiles = dlDirectory.listFiles();
    int fileNumber = 0;
    while(fileNumber < checkFiles.length)
    {
      // filename is after last slash; date is after first dash
      String checkFilePathname = checkFiles[fileNumber++].getName();
      String checkFilename = 
        checkFilePathname.substring(checkFilePathname.lastIndexOf('/')+1);
      int startDate = checkFilename.indexOf('-') + 1;
      startDate = checkFilename.indexOf('-',startDate) + 1;
      
      // if the date matches todays date we can stop 
      if(checkFilename.substring(startDate,startDate+8).equals(todaysDate) &&
         !checkFilename.endsWith("s.nwr"))
        return;
    }
    
    // we didn't find a file with todays's date, so copy the one we just made
    
    // make sure the directory configured is there
    File copyToDirectory = new File(recordingsDownloadDirectory);
    if(!copyToDirectory.exists())
    {
      System.err.println(
        "*** can't find directory to receive copy:" +
        recordingsDownloadDirectory);
      return;
    }
    if(!copyToDirectory.isDirectory())
    {
      System.err.println(
        "*** configured copy to directory not a directory:" +
        recordingsDownloadDirectory);
      return;
    }
    
    // copy the file
    if((new File(serverRecordingFileName)).length() < MIN_RECORDFILE_KEEP_SIZE)
    {
      System.out.println("\nno client recording uploaded but server recording " +
                         serverRecordingFileName + 
                         " not moved to downloads because it is under " +
                         MIN_RECORDFILE_KEEP_SIZE + " bytes");
      return;
    }
    String sourceFileName = new String(serverRecordingFileName);
    String copyFilename = 
      serverRecordingFileName.substring(serverRecordingFileName.lastIndexOf('/')+1);
    serverRecordingFileName = "";
    String copyfilePathname = recordingsDownloadDirectory + '/' + copyFilename;
    System.out.println("\ncopying serverside recording:" + sourceFileName +
                       "\nto Moodle download file:" + copyfilePathname);
    RecordingFileCopyThread recordingFileCopyThread = 
      new RecordingFileCopyThread(sourceFileName,
                                  copyfilePathname);
    runFileCopy = new Thread(recordingFileCopyThread);
    runFileCopy.start();
    
  }// end shutdownRecorder()
  
  // timestamp and queue messages to be sent to recorder
  // controlCode is 1 for control port, 0 for data
  // appNumber is 0 for audio, 1 for WBD, 2 for video, etc.
  public void queueForRecorder(byte[] message,
                               int controlCode,
                               int appNumber,
                               int udpMessageLength)
  {
	  // stop now if recorder is not working
	  if(!recordOnServer || terminateRecorder)return;
	
    // make timestamp byte array
	  byte[] timestamp = 
	  intToBytes((int)(System.currentTimeMillis()-recordingStartTime));
	
	  // allocate message and load header portion
	  byte[] newQueueMessage = new byte[udpMessageLength+9];
	  System.arraycopy(timestamp,0,newQueueMessage,0,4);
	  newQueueMessage[4] = intToByte1(udpMessageLength);
	  newQueueMessage[5] = intToByte2(udpMessageLength);
	  newQueueMessage[6] = (byte)controlCode;
	  newQueueMessage[7] = (byte)appNumber;
	  newQueueMessage[8] = 0;
    	
	  // copy the input message into queue message and add to queue
	  // (drop 5-byte TCP/UDP tunnel header)
	  System.arraycopy(message,5,newQueueMessage,9,udpMessageLength);
	  RecordQueueElement newQueueElement = 
	    new RecordQueueElement(newQueueMessage);
	  synchronized(recorderThread)
	  {
      if(recordQueue == null)recordQueue = newQueueElement;
	    else
	    {
	      RecordQueueElement chainToEnd = recordQueue;
	      while(chainToEnd.nextInQueue != null)
	        chainToEnd = chainToEnd.nextInQueue;
	      chainToEnd.nextInQueue = newQueueElement;
	    }
	  }

	  // notify the send thread that a message has been queued
	  synchronized(recorderThread)
	  {
	    recorderThread.notify();
	  }
  }// end queueforRecorder method
  
  
  // internal class to queue messages for recording
  class RecordQueueElement
  {
    byte[] queueMessage;
	  RecordQueueElement nextInQueue = null;
	
	  public RecordQueueElement(byte[] messageToQueue)
	  {
	    queueMessage = messageToQueue;
	  }
  }// end class RecordQueueElement


  // internal Thread class
  // This class is closely related to REC which provides a network-connected
  // recorder, but it lacks external control and network connections.
  // loop forever:
  //   pick up next message to record, from queue
  //   write it to file
  //   quit when recorder is shutdown
  class RecorderThread extends Thread
  {
	  public FileOutputStream outFile = null;
 
    public void run()
    {
      // make a unique output filename
      GregorianCalendar gcalendar = new GregorianCalendar();
	    int year = gcalendar.get(Calendar.YEAR)%100;
      int month = gcalendar.get(Calendar.MONTH) + 1 ;
      int date = gcalendar.get(Calendar.DATE);
      int hour = gcalendar.get(Calendar.HOUR_OF_DAY);
      int minute = gcalendar.get(Calendar.MINUTE);
	    String stryear = "" + ( year < 10 ? "0" + year : "" + year);
      String strmonth = "" + ( month < 10 ? "0" + month : "" + month);
      String strdate = "" + ( date < 10 ? "0" + date : "" + date);
      String strhour = "" + ( hour < 10 ? "0" + hour : "" + hour );
      String strminute = "" + ( minute < 10 ? "0" + minute : "" + minute );
      
      // make sure the directory exists
      recordingsSubdirectory = recordingsDirectory + "/" + moodleFileName;
      File testFile = new File(recordingsSubdirectory);
      if(!testFile.exists())testFile.mkdir();
      
      // we use moodleFileName both for recordings subdirectory and as part of filename
      serverRecordingFileName =  recordingsSubdirectory + "/" + moodleFileName + "-" + 
                                 stryear + "." + strmonth + "." + strdate + "-" + 
                                 strhour  + strminute + "s.nwr";
	  
	    // create the recording file
      try
      {
        outFile = new FileOutputStream(serverRecordingFileName);
      }
      catch(IOException ioe)
      {
        System.out.println("RecorderThread exception opening output file " + 
                           outFile + ":" + ioe.getMessage());
        return;
      }
	  
	    // log and flag recorder running
      terminateRecorder = false;
      System.out.println("MISTC Server recording class in file:" + 
                         serverRecordingFileName);
      
	    // continuously write data as it's available in the queue
      while(!terminateRecorder)
	    {
        // wait for a new message to be queued, then dequeue it
        synchronized(recorderThread)
        {
	        try
		      {
	          if(recordQueue == null)
			      {
			        wait();
			      }
          }
		      catch (InterruptedException ie)
		      {
		        terminateRecorder = true;
		      }
        }// end synchronized(queueSynch)
        
		    // write the message to outFile and advance to next
        if(recordQueue != null)
        {
		      synchronized(recorderThread)
		      {
		        try
		        {
		          outFile.write(recordQueue.queueMessage);
		        }
		        catch(IOException ioe)
		        {
		          System.out.println("Exception terminated RecorderThread:" + ioe);
              terminateRecorder = true;
		          break;
		        }
		        recordQueue = recordQueue.nextInQueue;
          }// end synchronized(recorderThread)
        }//end if(recordQueue != null)
		  
        // check for shutdown state
        if(terminateRecorder && (recordQueue == null))
        {
		      try
		      {
		        outFile.close();
          }
		      catch(IOException ioe)
		      {
		        System.out.println("Exception closing file in RecordThread:"+ioe);
          }
        }// end if(terminateRecorder...
      }// end while(!terminateRecorder)
          
      // check for zero-length recording; if found, delete
      testFile = new File(serverRecordingFileName);
      if(testFile.length() < MIN_RECORDFILE_KEEP_SIZE)
      {
        System.out.println("deleting server recording file " + serverRecordingFileName +
                           " (length under " + MIN_RECORDFILE_KEEP_SIZE + ")");
        testFile.delete();
      }
      return;

    }// end run()
    
  } // end RecorderThread class
  
  
  // internal Thread subclass to copy a recording file
  // parameters are path-filenames
  class RecordingFileCopyThread extends Thread
  {
    FileInputStream inFile = null;
    FileOutputStream outFile = null;
    String inFilePathname;
    String outFilePathname;
    byte[] header = new byte[9];
    byte[] dataBuffer = new byte[1500];
    int bytesCopied = 0;
    int headerBytesReadThisRecord = -1;
    int dataLengthFromHeader = -1;
    int dataBytesReadThisRecord = -1;
    boolean copyComplete = false;
    
    public RecordingFileCopyThread(String inFilePathnameRef,
                                   String outFilePathnameRef)
    {
      inFilePathname = new String(inFilePathnameRef);
      outFilePathname = new String(outFilePathnameRef);
      
    }// end constructor
    
    public void run()
    {
      // wait for the server RecorderThread to terminate
      try
      {
        while(!terminateRecorder)Thread.sleep(100);
      }
      catch(InterruptedException ie){}
      
      // open files
      try
      {
        inFile = new FileInputStream(inFilePathname);
      }
      catch(FileNotFoundException fnf)
      {
        System.err.println(
          "*** recording file not found to copy:" +
          inFilePathname);
      }
      try
      {
        outFile = new FileOutputStream(outFilePathname);
      }
      catch(IOException ioe)
      {
        System.err.println(
          "*** could not create recording file copy:" +
          outFilePathname);
      }
      if((inFile == null) || (outFile == null))return;
      int copySize = (int)(new File(inFilePathname)).length();
      
      // copy input to output, one recording segment at a time
      while(!stopTLM && (bytesCopied < copySize))
      {
        try
        {
          // check for complete header
          headerBytesReadThisRecord = inFile.read(header,0,9);
          if(headerBytesReadThisRecord != 9)break;
        }
        catch(IOException ioe)
        {
          headerBytesReadThisRecord = -1;
          break;
        }
        
        try
        {
          // get dataLength from header and check its size
          dataLengthFromHeader = bytesToUshort(header[4],header[5]);
          if(dataLengthFromHeader > dataBuffer.length)
          {
            System.err.println(
              "*** corrupted header containing length " + 
              dataLengthFromHeader +
              " read from file " + inFilePathname);
            dataLengthFromHeader = -1;
            break;
          }
          
          // write header to output
          outFile.write(header);
        
          // read the data
          dataBytesReadThisRecord = 
            inFile.read(dataBuffer,0,dataLengthFromHeader);
          if(dataBytesReadThisRecord != dataLengthFromHeader)
          {
            dataBytesReadThisRecord = -1;
            break;
          }
        
          // write data to output
          outFile.write(dataBuffer,0,dataLengthFromHeader);
        }
        catch(IOException ioe)
        {
          System.err.println("*** IOException:" + ioe);
          dataBytesReadThisRecord = -1;
          break;
        }
        
        // keep track of bytes processed
        bytesCopied += 9 + dataBytesReadThisRecord;
        
      }//end while(!stopTLM)
      
      // check for errors in copy
      if(headerBytesReadThisRecord != 9)
      {
        System.err.println(
          "*** corrupt header found in recordingFileCopy of file:" +
          inFilePathname + " after " + bytesCopied + " bytes");
          return;
      }
      if(dataBytesReadThisRecord != dataLengthFromHeader)
      {
        System.err.println(
          "*** correct data record found in recordingFileCopy of file:" +
          inFilePathname + " after " + bytesCopied + " bytes");
          return;
      }
      
      // close files and print stats
      try
      {
        inFile.close();
        outFile.close();
      }
      catch(IOException ioe)
      {
        System.err.println("*** IOException in file close:" + ioe);
        return;
      }
      System.out.println(
        "copied " + bytesCopied + " bytes from " + inFilePathname +
        " to " + outFilePathname);
         
    }// end run()
  
  }// end class RecordingFileCopyThread
  
} // end class ServerRecord
