/*
 * Transport Layer Multicaster main class
 * Created on August 31, 2001
 * @author  mpullen & students
 * @version 5.2.4
 * @modified JMP 3/1/13
 */
/*----------------------------------------------------------------*
|   Copyright 2001-2013 Networking and Simulation Laboratory      |
|         George Mason University, Fairfax, Virginia              |
|                                                                 |
| Permission to use, copy, modify, and distribute this            |
| software and its documentation for all purposes is hereby       |
| granted without fee, provided that the above copyright notice   |
| and this permission appear in all copies and in supporting      |
| documentation, and that the name of George Mason University     |
| not be used in advertising or publicity pertaining to           |
| distribution of the software without specific, written prior    |
| permission. GMU makes no representations about the suitability  |
| of this software for any purposes.  It is provided "AS IS"      |
| without express or implied warranties.  All risk associated     |
| with use of this software is expressly assumed by the user.     |
*----------------------------------------------------------------*/
/*Transport Layer Multicaster( TLM )v0.1 of NetworkEducationWare Project
   modified by Rajesh Adivi(vadivi@gmu.edu) FALL:2000(11/6/00)
   based on initial version by Vinod Mukkamala Spring 2000

 TLMv0.1 by Dilip Jeyram based on work of Rajesh Adivi and Vinod Mukkamala
  03/13/2001.
  Implemented: 'Packet send only after receive' for all unicast Hosts.
  Only If a host sends packets within the time duration w, will it receive any
  packets.

 TLMv0.2 by Dilip Jeyram following pseudocode by JMP

 TLMv1.0 by J.M.Pullen and Manjeera Parupalli (mparupal@gmu.edu) added
 floor control.

 TLMv1.1, 1.2, 1.3, 1.4, 1.5, 1.6, 1.7, 1.8, 2.0, 2.1, 3.0, 3.1, 3.2, 3.3 by JMP
 (mpullen@gmu.edu) evolved TLM to meet expanded definition of NEW with three
 major tools (audio, whiteboard, video) requiring prioritized switching

 TLMv4.0 by JMP and Charles Snow (CMS) added privacy, XML config, and scheduling

 TLMv4.1 by JMP added features to work with NAT and proxy: authentication
 without IP address and indentification of TCP connection by IP address plus
 source port

 TLMv4.2 by JMP added semester code to .conf file; fixed a race bug in
 TCP connections at login; sychronized floor access

 TLMv4.3 by JMP moved TLMFC connection to TLMC tunnel so TLMFC can use
 multicast/UDP connection under IRMP

 TLMv4.4 by JMP implemented auto-download of application zipfiles
 NOTE: this works only with TCP and download happens only while the
 floorholder is sending audio (which is almost always, in practice)
 
 TLMv4.4.4 by JMP added server-side recording

 TLMv4.5.0 by JMP added multicast tunnel, breakout rooms, and voting
 
 TLMv4.5.1 by JMP allows auto-reconnect
 
 TLMv5.0.0 by JMP supports MISTC
 
 TLMv5.1.0 by JMP maintenance release
 
 TLMv5.2.0 by JMP added server recording download to MISTC client
 
 TLMv5.2.4 by JMP supports auto-upload of recordings
***************************************************************************
Server-side recording and auto-upload of recordings:

1. Primary recordings under MISTC are made by client. This allows user
to determine what parts of session will be recorded, and to record when
disconnected from the server. But it is subject to human forgetfulness. 
2. If <recordOnServer> is set in conf file, the MISTC server created a .nwr
recording file, compatible with the client recording.
3. Transferring the recording files into Moodle cna be accomplished using
Web browser upload. However the process has been automated: whenever client
recording is stopped (not paused), client offers to upload that recording to 
the server. Server places it in recordingsDownloadDirectory, in a subdirectory
with name = number CID.
4. If no client recording is uploaded before the end of a session (as 
detected by the date code contained in the recording name) the server will
copy its serverside recording to the downloads directory that is exposed 
through Moodle. End of session is detected when last participant disconnects.
 
***************************************************************************
 NOTE on auto-scheduling: if indicated in the configuration file, TLM
 will control session participants by times set in the database table
 'courseinfo'. During periods when no course is scheduled, it is possible
 for a user from any course scheduled for this room to login to the server.
 During periods when a course is scheduled, only users with a 'userinfo'
 table entry that has 'courseID' matching that in the database table
 'courselist' can login. As an exception to this, any user coded
 'instructor' level 2 and 3 in database table 'userinfo' can login at
 any time. Upon transition from unscheduled time to a scheduled course,
 all users are automatically logged out. Auto-scheduling allows multiple
 courseIDs to share the server, with the available period set by the first
 one found in 'courselist'. If the 'courselist' schedule is changed, the
 server must be restarted to ensure the change will take effect.
***************************************************************************

 command line parameters to invoke TLM:
 1. configuration file path/name
    optional: defaults to TLM.conf
 2. semester code
    optional; defaults to null
    NOTE: the semester code will be taken from the database, if a database
    is defined in the configuration file and it contains a table "semesterinfo"
    that has a row with value "current" in the "conference" column. Otherwise
    command line parameter 2 will be used, if present
NOTE: The command line was changed effective with version 4.1; information
previously here is now provided through the configuration file.
**********************SYSTEM LOGIC:*************************   version of 11/6/01
TLM supports a mix of TCP and UDP connections for data, with TCP
for floor control and record/playback control. Client UDP ports are
expected to be assigned in pairs, where portNumber+1 is the control for
portNumber. After authentication/connection to floor control and option
TCP tunnel connection, the server waits to send data for each application
until some traffic (data or control) for that application is received
from the client. If received through TCP tunnel, that application will
be supported by TCP; if received through UDP, the application will be
supported by UDP. Both sets of streams are floor-controlled.

TLM.java ( main class ) logic:

->Read file TLM.conf for number of applications, tunnelTCPPort - port on which TCP
  connections are to be made, record/playback port, client and server addresses,
  multicast address/port, unicast addresses/port/authentication and privileges info,
  for the applications of conferencing/teaching tools. More config parameters
  below.
->Read in the appDownloadZipfile
->Open multicast and unicast UDP sockets
->Create two instances per application of MCListen and UCListen, for port and port+1;
  pass arguments multicast address, unicast address array, TCP tunnel port, multicast
  port,unicast port, multicast socket, unicast socket, multicast last recv time,
  unicast last recv time array, unicast TCP array (initially zero), and window w
->Spawn the FloorControlConnect thread which listens to tunnelTCPPort+1 for Floor Control
  connnections. Each of these spawns a FloorControlListen thread which accepts
  authentication, sets appropriate flags, then interacts. The flags are checked
  before sending a UDP message or establishing a TCP data connection in UCListen
  MCListen and TCPTunnel.
->Use NEWSocketChannel to listen for connections made by remote hosts on the tunnelTCPPort.
  When a remote host sends data on the tunnelTCPPort, create a dedicated socket
  to that host and set tcpIsOpen for this host indicating that this host is sending
  data on the tunnelTCPPort. Spawn a thread, TCPTunnel, and pass to it the input stream,
  address of the remote host and tcpIsOpen parameters. A second connection from
  the same client will spawn a TCPVideo thread. The NEWSocketChannel class produces either
  a non-blocking TCP Socket or a non-blocking SSL socket depending on its third parameter.
->As client connections are made, start downloading the appDownloadZipfile as
  capacity allows (only for clients configured/tested for video and download)
********************************************************************************
 Auth.java class logic:
 ->If database is active, validate parameters in database
 ->Otherwise, validate parameters in .auth file
********************************************************************************
FCConnect class logic (in module TLM.java):
->Until done is signalled
-> loop forever
-> open a listening TCP server socket
-> accept connection from any host and spawn a FloorControlListen thread
-> end loop forever
 *******************************************************************************
FCListen.java class logic:
-> read one message: the authentication data handle, key and implicitly host address
-> if the host address is present in the hostAddr array
->   check the handle for match
->     if there is a key for that host it must match also
-> else if an address with value one and matching handle is present in the hostAddr array
->   also check if the entry has a key it must match too
-> if either of the above conditions holds, set "authenticated" bit for the host
-> send one-byte message to connecting client: 1 for authenticated, 0 for failed
-> if failed, disconnect
-> loop forever
->   listen for additional messages from TLMFC and respond on a case basis
->   if a URL message is received, forward it to the TCPOutStreams for all hosts
->   and also to all other FC.outToClient streams
-> end loop forever
********************************************************************************
MCListen.java class logic:
->In an infinite loop
->  read multicast UDP messages
->  update multicast last time recv
->  if multicast UDP message is not from the bridge machine
->  forward only if source is in unicast array
     (insert in array if a wildcard slot is available)
->  also, forward only if floorHolder or port is not floor controlled
->    send A copy to each unicast UDP address that has been active within time
        window unless port is associated with TCP on this host
->    and to each unicast TCP address/TCP tunnel port
        (after prepending unicast UDP port to message)
********************************************************************************
UCListen.java class logic:
->In an infinite loop
->  read unicast UDP messages; forward only if they are in unicast array
      (insert in array if a wildcard slot is available)
->  also, forward only if floorHolder or port is not floor controlled
->  update unicast last recv time array
->  if the multicast group has been a source of messages (any port) within time window
      send unicast-udp message to multicast group with a modified TTL(15)
->  send a copy of unicast UDP message to each unicast addresses that has been active
      within time window except the address it came from (source address)
      unless the application is associated with TCP on that host
->  and to each unicast TCP address/TCP tunnel port except the address it came from
      (after prepending unicast UDP port to message)
********************************************************************************
RPListen.java class logic:
->Listen for TCP connection by record control
 ->In an infinite loop
 ->  read record/playback control TCP messages
 ->  forward the control message via all connected TCP tunnels
 ->On disconnect, reset for another connection
 *******************************************************************************
TCPTunnel.java class logic:
->Until the TCP connection is closed
-> read TCP message
-> strip port from first two byte of message and locate it in multicast port array
   or unicast port array
   if it is in the array, capture the corresponding port (unicast if it is in
   the multicast array and vice versa)  and flag that port as TCP for this host;
   else go back to reading TCP
-> if a loopback time probe message is received, send it back to the source
-> if not loopback or URL, it should be forwarded in the following ways
   -> update unicast last recv time array
   -> if the multicast group has been a source of messages (any port) within time
      window send unicast-udp message to multicast group with a modified TTL(15)
   -> send a copy of unicast UDP message to each unicast address that has been active
      within time window except the address it came from (source address) and queue
      it to be sent to each TCP and UDP tunnel address except the address it came
      from (tunnel gets the original message with port number prepended and revised
      transaction code)
-> if there is a record/play control connection open and the message bears rpCPort,
   send a copy through the tunnel connection to the control
-> When the connection is closed, set corresponding unicast TCP array to 0 and return
 ******************************************************************************
TCPVideo.java class logic:
->Until the TCP connection is closed
-> read TCP message
-> strip port from first two byte of message and locate it in multicast port array
   or unicast port array
   if it is in the array, capture the corresponding port (unicast if it is in
   the multicast array and vice versa)  and flag that port as TCP for this host;
   else go back to reading TCP
-> if not loopback or URL, it should be forwarded in the following ways
   -> update unicast last recv time array
   -> if the multicast group has been a source of messages (any port) within time
      window send unicast-udp message to multicast group with a modified TTL(15)
   -> send a copy of unicast UDP message to each unicast address that has been active
      within time window except the address it came from (source address) and queue
      it to be sent to each TCP and UDP tunnel address except the address it came
      from (tunnel gets the original message with port number prepended and revised
      transaction code)
-> When the connection is closed, set corresponding unicast TCP array to 0 and return
*******************************************************************************
TunnelSend.java class logic:
->In an infinite loop
 ->  dequeue a unicast message associated with this outstream
     on basis of priority
 ->  send the message to the associated TCP outstream
 -> if send slot not full, send zipDownload or recordingDownload if available
********************************************************************************
In addition to the command line, if a MySQL database is used for authentication
and logging, the database connection information is kept in the configuration file.

The configuration file defines network and operational parameters, including
specific IP addresses to be served; the authentication file provides usernames
with a password and previleges for each. TLM will first try to access the
MySQL database using the second argument; if that is not successful it will
try to open an authentication file of that name.

Definition of XML configuration file for TLM: see comments in ParseConfig class.

Port number convention: Although our program logic does not require it except
where noted, configuring the NEW system is made much simpler by following a
convention for port numbers. WE have chosen port numbers such that, if
basePort is the first port number used (for both TCP and UDP), the following
apply:

TCP Ports:
  TLM_TCP_Port = basePort (implies TLM FloorControl port = basePort+1)
  TLM_UDP_Port = basePort
UDP Ports:
  MC_FC_Port = 62001 (implies basePort at MC_FC_Port -1)
  Audio_Port = 62004 (implies a control port at 62005)
  WBD_Port   = 62008 (implies a control port at 62009)
  WBD2_Port  = 62012 (implies a control port at 62013)
  Video_Port = 62016 (implies a control port at 62017)
  Unicast_Ports = Audio_Port,WBD_Port,Video_Port

Definition of authentication file for TLM (where database is not used):
  the filename is courseID.auth where courseID comes from the configuration
  file and defaults to "TLM.auth". This is the alternative to using a MySQL
  database. File consists of one line per user, fields separated by /
  (comments have # in first position):
  1. username - up to 8 characters
  2. password - up to 8 characters
  3. Unicast Address of user, may be DNS or IP address; if present, user must
     be at this IP address; if missing, user may be at any IP address
  4. drawGrid - 0 or 1 (draw a floor control grid - defaults to 0)
  5. controlPrivilege (user is floor controller - defaults to 0)
     0 means user with no control privilege
     1 means normal controller;
     2 means a super-controller who can input to WBD
       even though they do not hold the floor
  6. fullname - up to 40 characters
  Items 2 to 6 are optional, as are slashes unless needed, however at least one
  character must appear between any two slashes.
  File must end with a new line containing "end;"

Definition of database access file for TLM:
  This file contains a single line with two parameters, separated
  by slashes:
    1. MySQL database account
    2. MySQL database password
  The database name is passed to TLM as a command-line parameter and need not
  be protected from access. File  should have access privileges restricted 
  such that other system users cannot read it.

*******************************************************************************
Message formats:
************************************************
Messages from TLMFC to TLM
transaction codes 0 to 28, 101, 102, 107, 108, 109, 111, 112, 117

 connectMessage TLMFC to TLM
 {
  transaction byte[1] value 0
  messageLength unsigned short value 30 // 2 bytes
  username byte[8]
  password byte[8]
  courseid byte[11]
  semester byte[3]
 }

 text system panel TLMFC to TLM
  {
   transaction byte[1] value 1
   messageLength unsigned// 2 bytes
   text byte[255 max]
  }

  chat text to controller ("PROF") only TLMFC to TLM
  {
   transaction byte[1] value 2
   messageLength unsigned// 2 bytes
   text byte[255 max]
  }

  chat text to all TLMFC to TLM
  {
   transaction byte[1] value 3
   messageLength unsigned// 2 bytes
   text byte[255 max]
  }

  URL (outgoing) TLMFC to TLM
  {
   transaction byte[1] value 4
   messageLength unsigned short // 2 bytes
   text byte{200 max]
  }

  floor request with no request pending TLMFC to TLM (via TLMC tunnel)
  {
   transaction byte[1] value 5
   messageLength unsigned short value 8 // 2 bytes
   username byte[8]
  }

  floor transfer to pink grid box TLMFC to TLM (via TLMC tunnel)
  {
   transaction byte[1] value 6
   messageLength unsigned short value 8 // 2 bytes
   username byte[8]
  }
  NOTE: to grant click on user's box

  floor status TLMFC to TLM (via TLMC tunnel)
 {
  transaction byte[1] value 7
  messageLength unsigned short value 1 // 2 bytes
  status byte[1]
 }

 yield floor TLMFC to TLM (via TLMC tunnel)
 {
   transaction code byte[1] value 8
   messageLength unsigned short value 8 // 2 bytes
   username byte[8]
 }

  request fullname TLMFC to TLM (via TLMC tunnel)
  {
   transaction code byte[1] value 9
   messageLength unsigned short value 8 // 2 bytes
   username byte[8]
  }

  dropped participant TLMFC to TLM
  {
    transaction code byte[1] value 10
    messageLength unsigned short value 8 // 2 bytes
    username byte[8]
  }

  session variable TLMFC to TLM (via TLMC tunnel)
  {
    transaction code byte[1] value 11
    messageLength unsigned short value 18 // 2 bytes
    sessionVariable byte[64]
    tlmcTcpPort unsigned short // 2 bytes
  }

  withdraw floor request with request pending TLMFC to TLM (via TLMC tunnel)
  {
   transaction byte[1] value 12
   messageLength unsigned short value 8 // 2 bytes
   username byte[8]
  }

  appLaunch command (outgoing) TLMFC to TLM
  {
   transaction byte[1] value 13
   messageLength unsigned short // 2 bytes
   text byte[127 max]
  }

  download zipfile identification TLMFC to TLM
  {
   transaction byte[1] value 14
   messageLength unsigned short // 2 bytes
   zipfilename byte[max 24]
  }
  
  setup ACK TLMFC to TLM
  {
   transaction byte[1] value 18
   messageLength unisgned short value 0 // 2 bytes
  }

  heartbeat TLMFC to TLM
  {
   transaction code byte[1] value 19
   messageLength unsigned short value 0 // 2 bytes
  }
  
  open voting panel TLMFC to TLM to TLMFC
  (server ensures this comes from a controller)
  {
   transaction code byte[1] value 20
   messageLength short (2 bytes) value 10
   date byte[10] (formatted yyyy.mm.dd)
  }
  
  destroy vote panel TLMFC to TLM to TLMFC
  (server ensures this comes from a controller)
  {
   transaction code byte[1] value 21
   messageLength short (2 bytes) value 0
  }
  
  start vote TLMFC to TLM to TLMFC
  (server ensures this comes from a controller)
  {
   transaction code byte[1] value 22
   messageLength short (2 bytes) value 4 + propositionLength
   voteSequence unsigned short      // 2 bytes 
   propositionLength unsigned short // 2 bytes
   proposition text byte[0 to 1400]
  }
  
  cutoff vote TLMFC to TLM to TLMFC
  (server ensures this comes from a controller)
  {
   transcaction code byte[1] value 23
   messageLength short (2 bytes) value 6
   count yes unsigned short     // 2 bytes
   count no unsigned short      // 2 bytes
   count abstain unsigned short // 2 bytes
   count possible unsigned short// 2 bytes
  }
  
  submit vote TLMFC to TLM to TLMFC
  {
   transaction code byte[1] value 24
   messageLength short (2 bytes) value 2
   vote code byte[1] value 1=abstain 2=yes 3=no
   change from vote byte[1] value 0=none 1=abstain 2=yes 3=no
  }
  
  request breakout data TLMFC to TLM
  {
   transaction code byte[1] value 25
   messageLength short (2 bytes) value 1
   goToBreakoutCode byte[1] value 1 to go, 0 to leave
  }
  
  assign user to breakout group TLMFC to TLM
  {
    transaction code byte[1] value 26
    messageLength short (2 bytes) value 9
    group byte[1] binary 1 through 9; 0 for none
    username byte[8] chardata
  }
 
  participant left group TLM to TLMFC (via TLMC tunnel)
  {
    transaction byte[1] value 27
    messageLength unsigned short value 8 // 2 bytes
    username byte[8]
  }
  
  goto breakout mode TLMFC to TLM
  {
    transaction code byte[1] value 27
    messageLength short (2 bytes) value 2
	  goto status byte[1] binary; values 0 no breakout; 1 breakout
	  instructor status byte[1] binary; value 0 send to all; 1 to 9 group to join
  }
  
  set number of groups TLMFC to TLM
  {
    transaction acode byte[1] value 28
	  messageLength short (2 bytes) value 1
	  numberOfGroups byte[1] binary
  }
  
  request recordingFileList TLMFC to TLM
  {
    transaction code byte[1] value 101
    messageLength short (2 bytes) value 0 or 1
    placeholder byte[1] if present, indicates fileSize required
    (response with no filesize is TC 50; with filesize TC %%%% 
  }
  
  request download of recording TLMFC to TLM
  {
    transaction code byte[1] value 102
    messageLength short (2 bytes)
    filename byte[messageLength]
  }
  
  request upload of appDataZipFile TLMFC to TLM
  {
    transaction code byte[1] value 107
    messageLength short (2 bytes) value 36
    fileLength long (8 bytes)
    fileLastModified (8 bytes)
    fileMoodleShortName (20 bytes trailing blanks)
  }
  
  response to request upload of appDataZipFile TLM to TLMFC
  {
    transaction code byte[1] value 108
    messageLength short (2 byte) value 1
    response byte[1] value 0 -> no, already have that zipfile
                           1 -> yes, upload that zipfile
  }
  
  appZipFileSeqment TLMFC to TLM
  {
    transaction code byte[1] value 109
    messageLength short (2 byte) value 1400 or less
    data byte[messageLength]
  }

  execute zip TLMFC to TLMC
  {
    tranaction code byte[1] value 111
    messageLength short (2 byte)
    unzipCommand byte[messageLength]
  }
  
  agree to logout previous instance TLMFC to TLM
  {
    tranaction code byte[1] value 117
    messageLength short (2 byte) value 1
    agree byte[1] value 0 no 1 yes
  }
  
************************************************
 Messages from TLM to TLMFC
 transaction codes 29 to 63
 
 update FC grid TLM to TLMFC
 {
   transaction code byte[1] value 29
   messageLength short(2 byte) value n*8
   userName byte[8][n] encoding 8-char String
 }
 
 goto breakout mode TLM to TLMFC
 {
   transaction code byte[1] value 30
   messageLength short (2 bytes) value 2
   goto status byte[1] binary; values 0 no breakout; 1 breakout
   instructor status byte[1] binary; value 0 send to all; 1 to 9 group to join
 }   

 provide breakout data TLM to TLMFC
 {
   transaction code byte[1] value 31
   mwessageLength unigned short (2 bytes) value 3 + 60*numberOfUsers
   configuredGroups byte[1] // char; stored as groupNumber in database table USERGROUPS 
							// where username="--------"
   numberOfUsers unsigned short  // count of all users in database for this PrimaryCourseID
   userDataGroup [numberOfUsers] // limited to 29 if using multicast FC
   {
     username byte[8]       // user's login ID; key to userinfo table in database
     userFullname byte[40]  // from userinfo table in database
     userGroup byte[1]      // char '1' through '9'
   }
 }

 failed authentication TLM to TLMFC
 {
   transaction byte[1] value 32
   messageLength unsigned short value 0
 }

 initialization TLM to TLMFC
 {
  transaction byte[1] value 33
  messageLength unsigned short value 21+48*(userCount-1)// 2 bytes
  initialFloorStatus byte[1] 1=ALWAYS, 2=NEVER, 3=ASKME
  controlPrivilege byte[1] // of this user
  drawGrid byte[1],        // 1 to draw it
  maxFcUsers byte[1],      // users; prior to v5.0.0 was rows of 5
  floorHolderName byte[8],
  thisSessionUsers unsigned byte[1] // count of all users who have 
                                    // logged in since session start  
  thisUsername byte[8],
  thisUserFullname byte[40],
  if drawGrid == "1", [thisSessionusers-1] other participants
  {
   username byte[8]  (username[0] is thisUsername)
   fullname byte[40] 
  }
 }

  system message TLM to TLMFC (via TLMC tunnel)
  {
   transaction byte[1] value 34
   messageLength unsigned short // 2 bytes
   text byte[200 max]
  }

  chattext TLM to TLMFC (via TLMC tunnel)
  {
   transaction byte[1] value 35
   messageLength unsigned short // 2 bytes
   text byte[200 max]
  }

  URL TLM to TLMFC (via TLMC tunnel)
  {
   transaction byte[1] value 36
   messageLength unsigned // 2 bytes
   text byte[200 max]
  }

  floor request response TLM to TLMFC (via TLMC tunnel)
   this turns grid box for user who asked pink on all panels
   that have controlPrivilege
  {
   transaction byte[1] value 37
   messageLength unsigned short value 8 // 2 bytes
   username byte[8]
  }

  floor grant TLM to TLMFC (via TLMC tunnel)
   this turns grid box for user who received red on all panels
  {
   transaction byte[1] value 38
   messageLength unsigned short value 8 // 2 bytes
   username byte[8]
   userFullname byte[40]
  }

  participant status TLM to TLMFC (via TLMC tunnel)
  {
   transaction byte[1] value 39
   messageLength unsigned short value 9 // 2 bytes
   thisIsTheController byte[1]
   username byte[8]
   fullname byte[40]
   downloadingStatus byte[1] 0->new; 1->downloading 2->can't download 3-> done downloading
  }

  dropped participant TLM to TLMFC (via TLMC tunnel)
  {
   transaction byte[1] value 40
   messageLength unsigned short value 8 // 2 bytes
   username byte[8]
  }

  floor control rules change TLM to TLMFC (via TLMC tunnel)
  {
   transaction code byte[1] value 41
   messageLength unsigned short value 1 // 2 bytes
   rule code byte[1] values
     1 - always
     2 - never
     3 - ask me
  }

  sending without being floor holder error TLM to TLMFC (via TLMC tunnel)
  {
    transaction code byte[1] value 42
    messageLength unsigned short value 0 // 2 bytes
  }

 requesting floor when you are floor holder error TLM to TLMFC (via TLMC tunnel)
 {
  transaction code byte[1] value 43
  messageLength unsigned short value 0 // 2 bytes
 }

 floor not available TLM to TLMFC (via TLMC tunnel)
  {
   transaction byte[1] value 44
   messageLength unsigned short value 8 // 2 bytes
   username byte[8]
  }

 granting floor when no request pending error TLM to TLMFC (via TLM tunnel)
 {
  transaction code byte[1] value 45
  messageLength unsigned short value 0 // 2 bytes
 }

 floor request registered TLM to TLMFC (via TLMC tunnel)
 {
  transaction code byte[1] value 46
  messageLength unsigned short value 0 // 2 bytes
 }

 setup to launch applications TLM to TLMC
 {
  transaction code byte[1] value 47
  messageLength unsigned short value 0 // 2 bytes
  byte[255 max] client application data path
 }

 terminate operation TLM to TLMFC (via TLMC tunnel or MC)
 {
  transaction code byte[1] value 48
  messageLength unsigned short value 0 // 2 bytes
 }

 system message TLM to TLMFC with no bell (via TLMC tunnel)
 {
  transaction byte[1] value 49
  messageLength unsigned short // 2 bytes
  text byte[200 max]
 }

 heartbeat TLM to TLMFC - no longer used (was code 50)
 
 downloadRecordingFilenames TLM to TLMFC
 {
  transaction code byte[1] value 50
  messageLength unsigned short value 0 // 2 bytes
  for each, filesize 8 chars + filename up to 30 chars 
    separated by / and catenated (if requesting TC 101
    has length 0, filesize is omitted)
 }

 textToController TLM to TLMFC (via TLMC tunnel)
 {
  transaction byte[1] value 51
  messageLength unsigned short // 2 bytes
  text byte[127 max]
 }

 invalid URL TLM to TLMFC (via TLMC tunnel)
 {
  transaction byte[1] value 52
  messageLength unsigned short value 0 // 2 bytes
 }

  reply fullname TLM to TLMFC (via TLMC tunnel)
  {
    transaction byte[1] value 53
    messageLength unsigned short value 48 // 2 bytes
    username byte[8]
    fullname byte[40]
  }

  invitation to identify download zipfile
  {
	  transaction byte[1] value 54
	  messageLength unsigned short value 0 // 2 bytes
	  invite/decline byte[1]
	    value 0: previously offered zipfile not available
	    all other values: server will accept a zipfilename
	  filler byte[1] value 0
	  fileList byte[up to 1500] separator "/" (empty for decline)

	  this message stimulates a transaction code 13 file stat
	  message in response; if fileList is null (messageLength = 0)
	  it requires no other output; if fileList has content
	  it also causes a popup fileselect on FC
  }

  withdraw floor request TLM to TLMFC (via TLMC tunnel)
  {
    transaction code byte[1] value 55
    messageLength unsigned short value 8 // 2 bytes
    username byte[8]
  }

  invalid session variable TLM to TLMFC (via UDP)
  {
    transaction byte[1] value 56
    messageLength unsigned short value 0 // 2 bytes
  }

  can't reach NEW clients TLM to TLMFC (via TLMC tunnel)
  {
   transaction byte[1] value 57
   messageLength unsigned short value 0 // 2 bytes
  }

  username already logged in TLM to TLMFC (via TLMC tunnel)
  {
   transaction byte[1] value 58
   messageLength unsigned short value 0 // 2 bytes
  }

  system full can't login TLM to TLMFC (via TLMC tunnel)
  {
   transaction byte[1] value 59
   messageLength unsigned short value 0 // 2 bytes
  }

  server disconnecting (no reason given) TLM to TLMFC (via TLMC tunnel)
  {
   transaction byte[1] value 60
   messageLength unsigned short value 0 // 2 bytes
  }

  failed authentication 3 times disconnected TLM to TLMFC (via TLMC tunnel)
  {
   transaction byte[1] value 61
   messageLength unsigned short value 0 // 2 bytes
  }

  server received unexpected message TLM to TLMFC (via TLMC tunnel)
  {
   transaction byte[1] value 62
   messageLength unsigned short value 0 // 2 bytes
  }

  appLaunch command (incoming) TLM to TLMFC
  {
   transaction byte[1] value 63
   messageLength unsigned short // 2 bytes
   text byte[200 max]
  }
  
  connect will disconnect previous instance TLM to TMFC
  {
   transaciton byte[1] value 118
   messageLength unsigned short value 0 // 2 bytes
  }

********************************************************
Messages between TLMC and TLM
transaction codes 64 to 95, 103, 104, 105, 106

 client startup TLM to TLMC
 {
  transaction byte value 64
  messageLength unsigned short value 25// 2 bytes
  clientVersionByte1          // byte
  clientVersionByte2          // byte
  clientVersionByte3          // byte
  connectionThreshold u_short // 2 bytes
  congestionThreshold u_short // 2 bytes
  recorderTCPPort u_short     // 2 bytes
  probeInterval unsigned short// 2 bytes
  appCount byte
  whiteboardRingSize unsigned short // 2 bytes
  roomID byte[11]             // MISTC-ID
  
  NOTE: Prior to v5.0.0 messageLength was 20 and in place of roomID was:
  semesterCode byte[3] // e.g. S07, or 000
  courseNum byte[5]    // unique course ID from database, or 00000
 }

 client startup TLMC to TLM
 {
  transaction byte[1] value 64
  messageLength unsigned short value 71 // 2 bytes
  clientVersionByte1 // byte
  clientVersionByte2 // byte
  clientVersionByte3 // byte
  sessionVariable byte[64]
  otherEndPort unsigned short    // 2 bytes
  netCapacitykbps unsigned short // 2 bytes
                                 // (short)0 for main reconnect
                                 // (short)1 for video reconnect 
 }

 client error message TLMC to TLM
 {
  transaction byte[1] value 65
  messageLength unsigned short (always 6)
  TLM error code byte
  errorInfo unsigned short // 2 bytes
 }

TLM error code values:
 1 received unexpected message format TLM to TLMC (errorInfo = first byte of received header)
 2 TCP/UDP message length mismatch TLM to TLMC (errorInfo = length from received header)
 3 error in UDP send to client (errorInfo = UDP port number)
 4 error in TCP receive from TLM (errorInfo = error code from TCP)
 5 error in UDP receive from client (errorInfo = UDP port number)
 6 error in TCP send to TLM (errorInfo = error code from TCP)
 7 received unexpected message format RP to TLMC (errorInfo = first byte of received header)
 8 TCP/UDP message length mismatch RP to TLMC (errorInfo = length from received header)
 9 bad UDP port in report/playback stream (errorInfo = UDP port number)
10 error in TCP send to RP (errorInfo = error code from TCP)
11 multicaster UDP message length received does not match internal length field
   (errorInfo = length from received header)
12 recorder UDP message length received exceeds max allowable size
   (errorInfo = length from received header)
13 UDP message length in header does not match received datagram length
   (errorInfo = length from received header)
14 TCP video tunnel received unexpected message format TLM to TLMC
   (errorInfo = first byte of received header)
15 download message received on unconfigured TLMC
16 download received segment out of sequence or bad segment length
17 error closing or unzipping download file

 client data message TLM to TLMC
 {
  transaction byte[1] value 66
  messageLength unsigned short
  portNumber unsigned short // zero for encapsulated floor control
  data byte [1 to 1500]
 }

 client data message TLMC to TLM
 {
  transaction byte[1] value 67
  messageLength unsigned short
  portNumber unsigned short // zero for encapsulated floor control
  data byte [1 to 1500]
 }

  client heartbeat message TLMC to TLM
 {
  transaction byte[1] value 68
  messageLength unsigned short value 1 // 2 bytes
  shutdownCode byte[1] value "1" to disconnect, "0" otherwise
  nullData byte[2] value "00"
 }

  audio-send message TLM to TLMC
 {
  transaction byte[1] value 69
  messageLength unsigned short value 0 // 2 bytes
  actionCode byte[1] value 0 to stop sending
                           1 to start
                           2 to stop with message
                             "you don't have floor"
                           3 to kill client with message
                             "login timed out"
  nulldata value byte[1] 0
 }

 loopback request TLMC to TLM
 {
  transaction byte[1] value 70
  messageLength unsigned short min 0 max MAX_MESSAGE_LENGTH // 2 bytes
  TLMCreportedRTT (ms) unsigned short // 2 bytes
  sequence number byte[1] // 0 or 1 using ABP
  payload  byte[messageLength-1]
 }

 URL for recorder TLM to TLMC
 URL from recorder TLMC to TLM
 {
  transaction byte[1] value 71
  urlLength unsigned short min 0 max 200 (if < 4 => just kill off URL)// 2 bytes
  notUsed byte[1] value 0
  urlCode byte[1] value 127
  url byte[urlLength]floor
 }

 no tunnel floor control TLMC to TLM
 {
   transaction byte[1] value 72
   messageLength unsigned short value 0// 2 bytes
   portNumber unsigned short value 0// 2 bytes
 }

 appDataZipFile download segment TLM to TLMC
 {
  transaction byte[1] value 73
  messageLength unsigned short // 2bytes 
  (zero messageLength indicates the whole file has been sent)
  segment SN unsigned short, mod 10000//2 bytes
  (add 10,000 to SN to request ACK)
  segment data byte[messageLength]
 }

 appDataZipFile name & unzip TLM to TLMC
 {
  transaction byte[1] value 74
  messageLength unsigned short value 27// 2 bytes
  unzip now byte[1] 1=yes 0=no
  change existing file date to mTime byte[1] 1=yes 0=no
  filename byte[19]
  mTime byte[4]
  length byte[4]
 }

 signal that unzip has succeeded TLMC to TLM & REC
 {
   transaction byte[1] value 75
   messageLength unsigned short value 0// 2 bytes
   nulldata byte[2] value "00"
 }

 signal successful end of network test TLMC to TLM
 {
  transaction byte[1] value 76
  messageLength unsigned short value 8// 2 bytes
  nulldata byte[2] value "00"
  appDataZipfileMtime unsigned int  // 4 bytes time_t format (seconds)
  // mTime is set to 0xffffffff if no file but download is supported
  appDataZipfileBytes unsigned int  // 4 bytes
 }

 signal unsuccessful end of network test TLMC to TLM
 {
  transaction byte[1] value 77
  messageLength unsigned short value 0// 2 bytes
  nulldata byte[2] value "00"
 }
 
 send filename for appending REC to RC
 {
  transaction byte[1] value 78
  messageLength unsigned short//2 bytes
  filename byte[messageLength-2]// minimum 2 bytes
 }
 
 ACK appFileDownload TLMC to TLM
 {
  transaction byte[1] value 79
  messageLength unsigned short value 0 // 2 bytes
  SNacked unsigned short
 }

 Messages between TLMC and record/playback servers
 transaction codes 80 to 87 and 113

 TLMC to record server
 {
  transaction code byte[1] value "80"
  timeStamp unsigned long (4 bytes)
  messageLength unsigned short (2 bytes)
  controlCode unsigned byte 0 or 1
  appNumber unsigned byte range 0 to number of Applications 
                          + 123/124/125/126
                          123: data is TC 74 appDataFilename
                               with two byets of control prepended
                          124: data is TC 73 appDataFileSegment
                               with 2 bytes of SN prepended
                          125: data is old-style appDataFileSegment
                          126: data is FCmessage
  data byte [1 to 1400]
 }

 playback server to TLMC
 {
  transaction code byte[1] value "81"
  messageLength unsigned short
  controlCode unsigned byte 0 or 1
  appNumber unsigned byte range 0 to number of Applications
  data byte [1 to 1400]
 }

 TLMC imbedded control message to record server
 {
  transaction code byte[1] value "82"
  messageLength unsigned short // 2 bytes
  controlCode byte[1]
  appNumber byte[1]
  control transaction code byte[1] value "74"
  control message byte[20]
 }

 record/playback server imbedded control message to TLMC
 {
  transaction code byte[1] value "83"
  messageLength unsigned short // 2 bytes
  controlCode byte[1]
  appNumber byte[1]
  control transaction code byte[1] value "75" or "77"
  control message byte[20]
 }

 record/playback appDataFile message (used with FC appLaunch record payload)
 {
  transaction code byte[1] value "84"
  messageLength unsigned short value "25"
  controlCode unsigned byte 0 or 1
  appNumber unsigned byte value 126
  unzip without download byte[1] 1=yes 0=no
  change existing file date to mTime byte[1] 1=yes 0=no
  filename byte[19]
  mTime byte[4]
 }

 record/playback appDataZipFile segment
 {
   transaction code byte[1] value "85"
   dataLength unsigned short // 2 bytes
   controlCode byte[1] value "0"
   appNumber byte[1] value "125"
   fileSegment byte[dataLength]
 }

 file list message between PLAY to PC
 {
  transaction code byte[1] value "88"
  messageLength unsigned short // 2 bytes = file list length in bytes, minus 2
  file list up to 40 files max 30 chars each delimited by "\"
 }
 
 file append message between REC and RC
 {
  transaction code byte[1] value "89"
  messageLength unsigned short // 2 bytes = file name length in bytes, minus 2
  file name up to 127 characters
 }
 
 file or directory not found PLAY to PC
 {
   transaction code byte[1] value "91"
   messageLength unsigned short value 0 // 2 bytes
 }
 
 download server recording message TLM to TLMC
 {
  transaction byte[1] value 103
  messageLength unsigned short // 2bytes 
  (zero messageLength indicates the whole file has been sent)
  segment SN unsigned short//2 bytes
  message header byte[9] // payloadLength short bytes 4-5
  payload data byte[payloadLength]
 }
 
 download server recording message ACK TLMC to TLM
 {
  transaction byte[1] value 103
  messageLength unsigned short value 0 //2 bytes
  segment SN unsigned short // 2 bytes
 }

 server recording name TLM to TLMC
 {
  transaction byte[1] value 104
  messageLength unsigned short value 38// 2 bytes
  NOTE: messageLength 0 informs TLMC that server can't download
  nulldata byte[2]
  filename byte[34]
  mTime byte[4]
 }
 
 reject last offered server recording TLMC to TLM
 {
  transaction byte[1] value 105
  messageLength unsigned short value 0 //2 bytes
  acceptCode byte[2](unsigned short) 0 to reject; other to accept
 }
 
 cancel ongoing recording download TLMC to TLM
 {
  transaction byte[1] value 106
  messageLength unsigned short value 0 // 2 bytes
  nulldata byte[2]
 }
 
 ACK appFileDownload TLMC to PLAY
 {
  transaction byte[1] value 113
  messageLength unsigned short value 0 // 2 bytes
  SNacked unsigned short
 } 
 
 request to server for REC to upload recording file REC to TLM via RC/FC/MISTC/TLMC
 {
  transaction byte[1] value 119
  messageLength unsigned short value filename length // 2 bytes
  filename byte[messageLength]
 }
 
 approval to upload recording file TLM to REC via TLMC/MISTC/FC/RC
 { 
  transaction byte[1] value 120
  messageLength unsigned short value 0 // 2 bytes
 }

 recording upload file segment REC to TLM via RC/FC/MISTC/TLMC
 {
  transaction code byte[1] value 121
  messageLength short (2 byte) value 1400 or less; zero at end
  data byte[messageLength]
 }
 
 recording upload file segment ACK TLM to REC via TLMC/MISTC/FC/RC
 {
  transaction code byte[1] value 121
  messageLength short (2 byte) value 2
  ACKNO short (2 byte) starts at 0; +1 per segment received
 }
 
 recording upload approval request REC to RC (for user display)
 {
  transaction code byte[1] value 122
  messageLength short (2 byte) value filename length max 255
  filename byte[filename length]
 }
 
 recording upload response RC to REC (based on user input)
 {
  transaction code byte[1] value 123
  messageLength short(2 byte) value 1
  data byte[1] value 0 (reject) or other (approve)
 }
 
 delete serverside recording FC to TLM via TLMC
   returned TLM to FC via TLMC as ACK; messageLength=0 is reject)
 {
  transaction code byte[1] value 124
  messageLength short (2 bytes) value filename length
  filename without path byte[filename length]
 }
 
 upload warning REC to TLMC via RC/FC/MISTC
 {
   transaction code byte[1] value 125
   messageLength short (2 bytes) value 1
   action code 1 = download starting 0 = download complete
 }
 
 text to display in System Messages REC to RC (and other?)
 {
  transaction code byte[1] value 126
  messageLength short (2 bytes) value 1 max 50
  data byte[messageLength]
 }
 
 filepathname for upload to server RC to REC
 {
   transaction code byte[1] value 127
   messageLength short (2 bytes) value filepathname length
   filepathname byte[messageLength]
 }
 
 *************************************************************************************
 messages between TLMC and TLMFC
 transaction codes 96 to 100, 110 and 115

 popup OK/CANCEL message TLMC to TLMFC
 {
   transaction byte[1] value 96
   messageLength unsigned short // 2 bytes
   messageText byte[messageLength} max length 511
 }

 reponse to OK/CANCEL TLMFC to TLMC
 {
   transaction byte[1] value 98
   messageLength unsigned short value 1 // 2 bytes
   responseCode byte[1] value 0 for OK, anything else for CANCEL
 }
 
 popup download OK/CANCEL message TLMC to TLMFC
 {
   transaction byte[1] value 99
   messageLength unsigned short // 2 bytes
   messageText byte[messageLength} max length 511
 }

 reponse to download OK/CANCEL TLMFC to TLMC
 {
   transaction byte[1] value 100
   messageLength unsigned short value 1 // 2 bytes
   responseCode byte[1] value 0 for OK, anything else for CANCEL
 }
 
 popup disconnect OK/CANCEL message TLMC to TLMFC
 {
   transaction byte[1] value 110
   messageLength unsigned short // 2 bytes
   messageText byte[messageLength} max length 511
 } 

 execute zip TLMFC to TLMC
 {
   tranaction code byte[1] value 111
   messageLength short (2 byte)
   zipCommand byte[messageLength]
 }

 zip complete TLMC to TLMFC
 {
   tranaction code byte[1] value 112
   messageLength short (2 byte) value 1
   success byte[1] value 0 succeeded 1 failed
 }
 
 send requestFloor TLMC to TLMFC
 {
   transaction code byte[1] value 114
   messageLength short (2 byte) value 0
 }
 
 shutdown breakoutPanel TLMC to TLMFC by way of tunnel 
 {
   transaction byte[1] value 115
   messageLength unsigned short value 0 // 2 bytes
 }
  
 new moodleShortTruncated TLMC to TLMFC
 {
   transaction code byte[1] value 116
   messageLength short (2 byte)
   moodleShortTruncated byte[messageLength]
 }

 
 *************************************************************************************/
//package TLM;
import java.io.*;
import java.net.*;
import java.util.*;
import java.lang.*;
import java.sql.*;
import java.nio.channels.SocketChannel;

class TLM
{
   // constants
   // the version codes are for lowest and highest versions of client we support
   static final byte[] minClientVersionBytes = {5,2,2}; // lowest version code
   static final byte[] maxClientVersionBytes = {5,2,4}; // highest version code
   static String minClientVersion = new String(minClientVersionBytes);
   static String maxClientVersion = new String(maxClientVersionBytes);
   static String patchLevel = "";
   static final String sixtyFourZeros =
     "0000000000000000000000000000000000000000000000000000000000000000";

   // NOTE: serverVersion need not match clientVersion; our convention is to
   // make new releases with matching numbers, but in many cases the server
   // will work with previous client versions. This must be determined by
   // the programmer and indicated in {min,max}ClientVersionBytes above.
   static final String serverVersion = "5.2.4"; // used for log printout
   
   static boolean stopTLM = false;                    // when true, shutting down session
   static boolean useSSL = false;                     // make TCP connections with SSL
   static final int authenticateTimes = 3;            // times user can try login before
                                                      // server disconnects
   static int fcDisconnectTime = 20000;               // milliseconds without FC heartbeat
                                                      // to disconnect
   static int tlmcDisconnectTime = 4000;             // milliseconds without TLMC message
                                                      // to disconnect
   static final double udpWindow = 30000;             // number of milliseconds within
                                                      // which a host must send UDP to
                                                      // receive the UDP distribution
   static final long reconnectTimeoutSec = 300;       // number of seconds to wait before clearing
                                                      // the host table after last user disconnects (we
                                                      // hold it uncleared to allow auto-reconnect)
   static final long reconnectTimeoutMillisec =       // previous line, in milliseconds
                     reconnectTimeoutSec*1000;
   static long normalReconnectTimeout = 10*1000;// 5 minutes;
   static long instructorReconnectTimeout = 3*60*60*1000;//3 hours
   static final int MAX_HOSTS = 25;                   // max hosts to be served at one time
   static final int MAX_BREAKOUT_GROUPS = 9;          // max groups in breakout mode
   static final int MAX_USERS = 300;                  // max users in authorization table for a session
   static final int MAX_MESSAGE = 1455;               // max UDP message payload for Ethernet
   static final int MAX_SCHEDULE_ENTRIES = 20;        // max schedule table size
   static final int MAX_CHAINED_SERVERS = 10;         // max chained server table size
   static final int MAX_COURSEID_ALIASES = 5;         // max courseID table size
   static final int DOWNLOAD_MESSAGE_SIZE = 1250;     // bytes in slice for application zipfile download
   static final int AUDIO_WINDOW_MS = 200;            // milliseconds in a download window
   static final int DOWNLOAD_MAX_SIZE = 10485760;     // largest appZipFile to download (10MB)
   static final int TUNNEL_ACTIVATION_TIMEOUT = 120;  // seconds we will wait for tunnel activation
   static final int TUNNEL_ACTIVATION_SLEEP_MS = 1000;// one second for tunnel activation wait
   static final int FC_SETUP_TIMEOUT = 240;           // quarter-seconds we will wait for FC setup
   static final int FC_SETUP_SLEEP_MS = 250;          // quarter second for FC setup sleep
                                                      // based on SN field and DOWNLOAD_MESSAGE_SIZE
   static final int MIN_RECORDFILE_KEEP_SIZE = 100000;// server record files below this size are deleted
                                                      // (100KB is about 1 minute elapsed recording time) 
   static String authFileName = null;                 // of for courseID.auth
   static String databaseHost = "localhost";          // host where database runs
   static String databaseName = null;                 // course management database
   static String databasePrefix = "";                 // Moodle database prefix
   static String databaseUsername = null;             // username to connect to database
   static String databasePassword = null;             // password to connect to database
   static String moodleShortName = null;              // short identifier Moodle associates with course/section
   static String moodleFileName = null;               // cleaned-up version of Moodle short name
   static boolean checkPassword = true;               // => use password authentication
   static boolean webAuthenticate;                    // => use Webserver authentication
   static int sleepTime = 100;                        // sleep duration for threads in ms
   static int maxLoginTime = 60000;                   // max time to login, in ms
   static int minimumRate = 25;                       // value in kbits/sec used to calculate
                                                      // connectionThreshold = 60000/minimumRate ms
                                                      // default 25 kb/s gives 2400 ms threshold and
                                                      // matches TCP capacity of 33 kb/s modem uplink
                                                      // associated with 56K modems
   static int connectionThreshold;                    // value in ms to trigger warning to user
                                                      // if exceeded from transmit of five
                                                      // segments of 10k payload bits each
   static int congestionThreshold = 1000;             // value in ms to trigger congestion control
                                                      // disconnects for congestion:
   static int audioCutoff = 100;                      // number of dropped audio messages to disconnect
   static int whiteboardCutoff = 200;                 // " whiteboard messages
   static int videoCutoff = 300;                      // " video messages
   static int multicastCutoff = 500;                  // " multicast messages
   static String configurationVersion;                // version code of config file (not of TLM)
   protected final static String endTag = "end;";
   protected final static String commentTag = "#";
   static String configFileName = null;
   static String startupDateTime;
   static String serverRecordingFileName = "";

   // udp and tcp sockets etc
   static MulticastSocket mcSocket1, mcSocket2, mcFcSocket;
   static MulticastSocket mcTunnelSendSocket, mcTunnelRecvSocket;
   static MulticastSocket[] mcSocket1Save = null, mcSocket2Save = null;
   static DatagramSocket udpSocket1, udpSocket2, tunnelMessageUDPSocket;
   static DatagramSocket[] udpSocket1Save, udpSocket2Save;
   static SocketChannel tcpRecvSocketChannel;
   static Object recvSocketLock = new Object();
   static SocketChannel[] tcpSocketChannels;
   static DatagramSocket[] udpSockets;
   static SocketChannel[] tcpVideoSocketChannels;
   static Thread[] sendThreads;
   static Thread[] videoThreads;
   static TunnelSend[] tunnelSends;
   static TCPHandoff[] tcpHandoffs;
   static Socket[] fcSockets;
   static MCFCListen mcfcListen;
   static IRMP irmp;
   static int tcpSocketReceiveBuffersize = 2*262143;// 2*sizeof largest WBD slide
   static int tcpSocketSendBuffersize = 2*262143;
   static int videoSocketReceiveBuffersize = 65536;
   static int videoSocketSendBuffersize = 65536;
   static BufferedOutputStream[] fcOutStreams;
   static BufferedInputStream[] fcInStreams;
   static boolean[] tunnelClosing;
   static int[] numAuthenticateTries;

   // control information
   static boolean usingMcFC    = false;  // indicates floor control can connect
                                         // via multicast
   static final int mcFCUserLimit = 29;  // number of users that can fit in TC 31 
                                         // message under UDP limit of 1455 bytes
                                         // (3 bytes plus 49 bytes per user)
   static boolean autoSchedule = false;  // indicates server is to be scheduled for
                                         // courses based on database 'courselist'
   static boolean hideFullname = false;  // indicates server is not to provide
                                         // full username in response to floor control
   static int probeInterval = 1000;      // value in ms of RTT probe repeat
   static boolean audioFloorControl;     // suppress non-floorHolder audio traffic
   static boolean wbdFloorControl;       // suppress non-floorHolder WBD actions
                                         // except auto request for retransmission
   static boolean videoFloorControl;     // suppress non-floorHolder video traffic
   static byte fcMaxUsers = 0;           // number of user rows in Floor Control
   static int maxConnections = 0;        // limit on concurrent connections-
                                         // to be filled in below
   static int audioPort = 0;             // audio port from config
   static int wbdPort = 0;               // whiteboard port from config
   static int videoPort = 0;             // video port from config
   static String videoAddress = "";      // for a "priveliged" video source outside FC
   static InetAddress videoInetAddress;  // IPAddress equivalent of videoAddress
   static boolean videoAddressConfigured = false;
   static boolean noDatabase = false;    // indicates no database found
   static int audioRingSize = 50;        // number of buffers in the audio buffer ring
   static int fcRingSize = 20;           // number of buffers in the floor control ring
   static int videoRingSize = 20;        // number of buffers in the video buffer ring
   static int whiteboardRingSize = 1000; // number of buffers in the whiteboard buffer ring
                                         // (needed for large individual slide files)
   static int multicastRingSize = 100;   // number of buffers in the multicast buffer ring
   static int dropMessageInterval = 100; // interval between warning messages
   static boolean retransmitUDP = false; // received UDP from LAN is to be retransmitted
   static int numberOfScheduleEntries;   // optional scheduling feature
   static int numberOfChainedServers;    // optional partnered servers
   static int numberOfReservedAddresses; // optional host table entries held for specific
                                         // configured address
   static String printCourseDayCode;     // Mon Tues etc for next course schedule
   static String printThisDayOfWeek;     // Mon Tues etc for today
   static long nextCourseStart;          // used to schedule connection of users in a course
   static String printStartTime;         // printable version of nextCourseStart
   static long nextCourseEnd;            // use to schedule disconnect of users in a course
   static String printEndTime;           // printable version of nextCourseEnd
   static String[] nextCourseID;         // all courseIDs associated with nextClassTime
   static String[] nextCourseNum;        // the courseNums associated with the courseIDs
   static int numberOfCourseIDAliases;   // count of entries in nextCourseID;
   static boolean classInSession = false;// true when a scheduled class is happening
   static int floorHolder;               // position in host table arrays of current
                                         // floor holder (-1 for nobody)
   static int previousFloorHolder;       // ditto for previous holder 
										                     // (-1 only until first user takes floor)
   static String floorHolderName;        // name associated with the floorHolder
   static Auth auth;                     // authentication class
   static FCConnect fcConnect;           // floor control connection manager
   static Runtime runTime = Runtime.getRuntime();
   
   // recording information (NOTE: recording runs under auto-schedule)
   static boolean recordOnServer = false;   // record auto-scheduled sessions on server
							                              // as backup in case client-side record fails
   static int recordOnServerConfig = 0;     // binary value from config file; turns on recording
   static ServerRecord serverRecord = null; // does the connect and forwarding
   static String recordingsDirectory = null;// where we will store recordings produced
   static String recordingsSubdirectory = null;// server recordings subdirectory for this course
   static boolean recordVideo = false;      // whether to record video traffic
   static boolean sendingSlide = false;     // indicates the floorholder WBD is sending
                                            // a slide and has not yet send draw_grp_ps
   static long slideTimestamp = 0;          // sending WBD's timestamp
   static long slidePageNumber = 0;         // sending WBD's page number
   static long slideIpAddress = 0;          // sending WBD's IP address
   static int fcSlideSendTimeout;           // ms to hold floor change when sending
   static Thread floorChangeHoldThread = null;
   
   // uploading and archiving recordings
   static String recordingsDownloadDirectory = null;
                                            // directory where we'll upload and copy recordings
                                            // for access in Moodle
   static String recordingUploadFilename = null;
                                            // has a value when we're uploading
   static UploadRecording[] thisHostUpload; // multiple hosts could be uploading at same time
   static Thread[] thisHostUploadThread;    // thread assocated with thisHostUpload
   static int[] recordDownloadMessagesSent;
   static int[] recordDownloadAcksReceived;
   
   // appLaunching
   static String appFilePath = "";          // path to directory holding appZipFile.zip
   static long appFileDownloadMtime;        // last modified time of appZipFile.zip time_t (s)
   static byte[] appFileDownloadMtimeBytes; // appFileDownloadMtime encoded in seconds
                                            // NOTE: we do millisec elsewhere
   static byte[] appFileDownloadLengthBytes;
   static int appFileDownloadLimit = 0;     // max bytes allowed in download zipfile
   static boolean downloadOfferedToController
                                = false;    // if not during a course time under autoschedule,
                                            // first controller to login gets chance to pick
                                            // download zipfile
   static int hostToWhichDownloadOffered;   // hostname when downloadOfferedToController was set
   static int appFileDownloadSize = 0;      // actual size of app file download as read
   static String appFileDownloadName ="";   // holds the name of zipfile we buffer and download
   static boolean downloadBuffering = false;//indicates downloadBuffer is running 
                                            //(download not ready)
   static boolean downloadIsReady = false;  //indicates download file has been buffered
   
   // downloading serverside recordings
   static long recordingFileDownloadSize = 0;// bytes to download
   static long recordingFileDownloadMtime;//last modified time of recording file
   static String recordingDownloadFilename = "";// path-name of file to download
   static boolean serverRecordingDownload = false;// when true we are downloading

   // floor control information
   // uncomment the next two lines to replace *** with only first controller's if
   // users in FC mode 3 ASK ME are to communicate before any controller logs in;
   // makes the software work as if controller has logged in and logged off
   // static boolean controllerLoggedIn = true;
   // static int controllerNumber = 1;
   static String thisCourseID = "";      // 11-char max identifier for course/conf
   static String thisCourseNum = "00000";// 5-char unique number for course/conf
   static byte[] thisCourseNumBytes = thisCourseNum.getBytes();
   static String roomID = "";            // 11-char max identifier for "room";
                                         // used when no course is scheduled
   static byte[] roomIDBytes;            // roomID stored in byte[]
   static int CID;                       // integer from .conf, used to make roomID
   static byte[] thisSemesterBytes;      // semester code in byte[]
   static String thisSemester = "   ";   // 3-char semester code for example F02
   static boolean controllerLoggedIn = false;
   static int numberOfControllersLoggedIn = 0;
   static byte floorStatus;              // 1 => ALWAYS,
                                         // 2 => NEVER(LOCKED),
                                         // 3 => ASK ME
   static byte initialFloorStatus = 1;   // same codes; keeps value to restore
                                         // at session end
   static boolean inBreakoutMode = false;// traffic is being distributed by room not to all
   static int breakoutController = -1;   // hostIndex for instructor initiating breakout mode
   static String breakoutControllerUsername; // always padded to 8 characters
   static byte instructorBreakoutGroup = 0;// group that instructor is sending to (0 = all groups)
   static byte oldInstructorBreakoutGroup = 0;
   static int numberOfBreakoutGroups = 0;// currently setup rooms
   static byte[] breakoutDataMessage = null;// formatted breakout message for late joiners
   static int[] breakoutGroup;           // number of group for each participant
   static int[] breakoutFloorHolder;     // in breakout mode, floorholder for each group
   static String[] breakoutFloorHolderName;// in breakout mode, username of floorholder
   static Object updateUserlist = new Object();
                                         // used to synchronize user changes in host table
   										 
   // per-user control state
   static boolean[] nameExistsAgain;     // indicates duplicate name in login process
   static boolean[] fcIsInitialized;     // true after setup messages have been acked by to FC
   static boolean[] fcIsReadyForAppData; // true after FC sends read message (TC 14)
   static int[] netCapacitykbps;         // net capacity passed from TLMC at setup
   static boolean[] reconnected;         // false for initial connection; true if immediate reconnect
   static int[] appFileMtime;            // last mod time of appDataZipFile passed from TLMC
   static int[] appFileSize;             // file size of appDataZipFile passed from TLMC
   static boolean[] appFileStatNew;      // set as semaphore when stat changes
   static boolean[] unzipIsReady;        // per-host indicator that download has been unzipped
   static byte[] appFileDownload = null; // holds the app data zipfile
   static boolean[] udpSignalSlideDone;  // per-sender signal flag
   static InetAddress[] hostInetAddr;    // unicast host addresses
   static String[] hostAddr;             // mirrors hostInetAddr
                                         // in format InetAddress.getHostAddress()
   static int[] sourcePort;              // port on sending host
   static int[] sourceVideoPort;         // video connection port on sending hsot
   static String[] sessionVariable;      // session variable passed from TLMC
   static int[] otherEndPort;            // port at other end of TLMC first connection
   static boolean[] udpActive;           // holds indices of hosts with active UDP flows
   static boolean[] anonymousHost;       // reflects status of hostAddr in like position;
                                         // false if address required; true if not
   static long[] uRecvTime;              // last TCP or UDP message for each UC host
   static long[] mRecvTime;              // last UDP message for each MC group
   static long[] fRecvTime;              // last floor control for each host
   static boolean[] fcIsMc;              // true if floor control is multicast
   static int[] loopbackMinRTT;          // RTT captured by loopback before any traffic
                                         // (only used for UDP tunnels)
   static int[] latestRTTIncrease;       // delta between minRTT and reportedRTT, by host
   static int[] audioDropped;            // messages dropped for non-video ports
   static int[] videoDropped;            // ditto for video port
   static int[] whiteboardDropped;       // ditto for all whiteboard ports
   static int[] multicastDropped;        // ditto for multicast port
   static byte[] lastErrorMessageType;   // last error message received from host
   static long[] lastWarningTime;        // when this host was warned by floor control
   static long[] udpSetupTime;           // time when last UDP setup message sent to host
   static int[] peakRTT;                 // largest reported RTT during life of a connection
   static byte[][][] audioQueues;        // buffers waiting transmit
   static byte[][][] fcQueues;           //
   static byte[][][] videoQueues;        // dimensions: [host][bufferRing][messageByte]
   static byte[][][] whiteboardQueues;
   static byte[][][] multicastQueues;
   static int[][] audioLengths;          // message lengths waiting transmit
   static int[][] fcLengths;
   static int[][] videoLengths;          // dimensions: [host][bufferRing]
   static int[][] whiteboardLengths;
   static int[][] multicastLengths;
   static int[] audioWindowStart, audioWindowEnd; // buffer windows in ring (circular)
   static int[] videoWindowStart, videoWindowEnd; // Start is first used position
   static int[] fcWindowStart,fcWindowEnd;
   static int[] whiteboardWindowStart, whiteboardWindowEnd; // End is one past last used
                                         // position (if Start==End the buffer is empty)
   static int[] multicastWindowStart, multicastWindowEnd;
   static boolean[] tcpIsOpen;
   static boolean[] connectedByTCP;
   static boolean[] connectedVideoByTCP;
   static boolean[] protocolIsTCP;
   static boolean[] connectionTestPassed;
   static TCPTunnel[] tcpTunnelInstance;
   static Thread[] tcpTunnelThreads;
   static int[] udpLoopbackCount;
   static boolean[] tunnelIsActive;
   static String[] username = null;      // 8-char unique identifier of authenticated user
   static String[] fullname = null;      // 40-char white pages name of authenticated user
   static String[] userCourseID = null;  // 11-char login identifier of course
   static long[] loginTime;              // holds system time at login
   static char[] controlPrivilege;       // 0 or 1 indicating whether each host
                                         // has floor control privilege; 2
                                         // for host that can add to WBD while
                                         // not holding the floor
   static int[] instructorPrivilege;     // 0 to 3 indicating whether each host
                                         // has course management privileges
                                         // 0=student, 1=teacher, 2=support, 3=admin
   static boolean[] drawGrid;            // 0 or 1 indicating whether client on this
                                         // host displays a floor control grid
   static boolean[] killConnection;      // normally 1; set to 0 when authentication
                                         // fails during tunnel setup process
   static int[] localFloorStatus;        // floor status as seen by each FC
   static byte[] openVoteDialog = null;  // message to open vote panel for late joiners
   static byte[] startAVote = null;      // message to start a vote for late joiners
   static int openVoteFloorholder = -1;  // number of host that opened vote panel
   static boolean[] setupReceived;       // indicates setup message has been received by tunnel
   static boolean[] authenticatedHost;
  
   // indicates that this host has been authenticated, by one of three mechanisms:
   // 1. 0 in IP address field allows anonymous connection (set when host connects)
   // 2. IP address with no username/password is automatically authenticated always
   // 3. if username (with or without password) is provided in config file,
   //    authenticated host will be set true if values match in separate authenticating
   //    connection from a TLMFC client
   
   // arrays of unicast and multicast ports, addresses; tunnelTCPPort
   static int[] udpPort = null;         // unicast UDP ports for each application
   static int[] mcPort = null;          // multicast UDP ports for each application
   static InetAddress[] mcGroup;        // multicast group addresses
   static int tunnelTcpPort = 0;        // for TCP tunnel connections from TLMC
   static int tunnelUdpPort = 0;        // for UDP tunnel from TLMC
   static InetAddress mcFcInetAddr = null;// for multicast floor control
   static int mcFcPort = 0;             // for multicast floor control
   static InetAddress multicastTunnelGroup = null;
                                        // address for multicast application tunnel
   static int multicastTunnelPort = 0;  // port number for multicast application tunnel
   static boolean multicastTunnelFloorControl = true;
                                        // limits multicast sending to floorholder
   static MCTunnelListen mctl;          // Thread class listens to local subnet multicast
   static InetAddress rpAddr = null;    // record/player address
   static InetAddress rpControlAddr = null;// record/player control address
   static int highestReservedHost;      // entries in host table below this are
                                        // reserved, not given out to users in auth table
   static InetAddress[] chainedServer;  // partner servers chained to this one

   // counters and buffers
   static int numberOfTunnelsActive = 0; // number of tcp connections to server
   static int udpFlowsActive = 0;        // number of UDP flows being forwarded
   static int tcpConnectionsMade = 0;    // total number since startup
   static int udpTunnelsMade = 0;        // "
   static int numberOfApplications = 0;  // total applications
   static boolean multicastIsConfigured = false; // supporting audio/WBD/video MC or not
   static boolean multicastTunnel = false; // supporting generic MC application tunnel
   static boolean multicastTunnelServerSubnet = false;
                                         // exchange tunnel multicast with server subnet
   static int numberOfHosts;             // occupies positions in host table
   static InetAddress localAddress = null; // multicaster host address
   static String localHostAddress;       // format InetAddress.getHostAddress()
   static InetAddress loopbackAddress;   // /127.0.0.1
   static String loopbackHostAddress;    // 127.0.0.1
   static long timeStart;                // time when TLM starts running
   static long timeEnd;                  // time when TLM is shut down
   static long countMcRecv;              // multicast messages received
   static long countMcSend;              // multicast messages sent
   static long countUdpRecv;             // unicast UDP messages received
   static long countUdpSend;             // unicast UDP messages sent
   static long countUdpTunnelRecv;       // UDP encapsulated in UDP tunnel received
   static long countUdpTunnelSend;       // UDP encapsulated in UDP tunnel sent
   static long countTcpRecv;             // UDP encapsulated in TCP tunnel received
   static long countTcpSend;             // UDP encapsulated in TCP tunnal sent
   static long countTime;                // total total time server active
   static byte[] shutdown = {68,0,1,1};  // shutdown advice to TLMC
     
  // class to hold information from a row in the database certifinfo table
  static class certificateInfo
  {
    String semester, courseID, username, commonName;
    String orgName, orgUnit, orgLoc, orgState, orgCountry, emailAddr;
    short downloadStatus = 0;
    long serialNum = 0;
    String createDate, downloadDate, expiryDate;

    String getSemester(){return semester;}
    String getcourseID(){return courseID;}
    String getUsername(){return username;}
    String getCommonName(){return commonName;}
    String getOrgName(){return orgName;}
    String getOrgUnit(){return orgUnit;}
    String getOrgLoc(){return orgLoc;}
    String getOrgState(){return orgState;}
    String getOrgCountry(){return orgCountry;}
    String getEmailAddr(){return emailAddr;}
    short getDownloadStatus(){return downloadStatus;}
    long getSerialNumber(){return serialNum;}
    String getCreateDate(){return createDate;}
    String getDownloadDate(){return downloadDate;}
    String getExpiryDate(){return expiryDate;}
  }

  // class to hold information about a user who logs in
  class loginData
  {
    String username = "        ";
    String courseID = "           ";
    String courseNum = "00000";
  }

  /***************************************** main ***********************************/
  public static void main( String[] args )
  {
    // initialize
    timeStart = System.currentTimeMillis();
    startupDateTime = dateTime();
    countMcRecv = 0;
    countMcSend = 0;
    countUdpRecv = 0;
    countUdpSend = 0;
    countUdpTunnelRecv = 0;
    countUdpTunnelSend = 0;
    countTcpRecv = 0;
    countTcpSend = 0;
    numberOfHosts = 0;
    floorHolder = -1;
	  previousFloorHolder = -1;
    floorHolderName = "********";
    multicastIsConfigured = false;
    nextCourseID = new String[MAX_COURSEID_ALIASES];
    nextCourseNum = new String[MAX_COURSEID_ALIASES];
    numberOfCourseIDAliases = 0;

    // read command line arguments; set up data structures accordingly
    try
    {
      localAddress = InetAddress.getLocalHost();
      localHostAddress = localAddress.getHostAddress();
      loopbackHostAddress = "127.0.0.1";
      byte[] loopback = {127,0,0,1};
      loopbackAddress = InetAddress.getByAddress(loopback);
    }
    catch(UnknownHostException uhe)
    {
      System.out.println("unable to determine local host Internet address");
      stopTLM = true;
      System.exit(1);
    }
    System.out.print("\nMISTC Server Transport Layer Multicaster version " +
                     serverVersion + " " + patchLevel + "\nrunning on:" +
                     localAddress + " started at:" + dateTime() + "\n");
    System.out.print("Compatible clients TLMC versions " +
                     minClientVersionBytes[0] + "." +
                     minClientVersionBytes[1] + "." +
                     minClientVersionBytes[2] + " to " +
                     maxClientVersionBytes[0] + "." +
                     maxClientVersionBytes[1] + "." +
                     maxClientVersionBytes[2] + "\n");
    System.out.print("George Mason University C4I Center Networking and Simulation Laboratory\n");
    System.out.print("copyright George Mason University 2001-2009\n");

    if( args.length < 1 )
    {
      // default configuration file name - " TLM.conf "
      configFileName = new String("TLM.conf");
      System.out.println( "default configuration file " +
                        "( TLM.conf ) is being used\n" );
    }
    else
    {
      // use the configuration file specified by the user
      configFileName = new String(args[0]);
      System.out.println("configuration taken from file:" + args[0]);
    }
    // test for existence of configuration file
    if(!(new File(configFileName)).exists())
    {
      System.out.println("configuration file not found");
      stopTLM = true;
      System.exit(1);
    }
    // pick up semester argument
    thisSemester = "   ";
    if( args.length > 1 )thisSemester = args[1];
    if(thisSemester.equals("   "))
      System.out.println(
        "semester code not configured in command line (maybe set by database or .conf file)");
    else
      System.out.println("using semester code:\"" + thisSemester + "\"");

    // parse configuration file
    if(ParseConfig.parseTlmConfig(configFileName))
    {
      stopTLM = true;
      System.out.println("TLM shut down.");
      return;
    }
    
    if(databasePrefix == null)
      moodleShortName = roomID;
    else
    {
      // get the Moodle ShortName for this course
      DBQuery dbq = null;
	    try
	    {
	      dbq = new DBQuery("jdbc:mysql://" + databaseHost + "/" + databaseName,
                          databaseUsername, databasePassword, databasePrefix);
        moodleShortName = dbq.getMoodleShortName(roomID);
	      moodleFileName = makeFileName(moodleShortName);
        if(moodleFileName.length() > 12)
          moodleFileName = moodleFileName.substring(0,12);
      }
	    catch(ClassNotFoundException cnf)
      {
        System.out.println("*** missing DBQuery class in TLM ");
        System.out.println("*** moodleShortName set to nullstring");
        moodleShortName = "";
        moodleFileName = "";
      }
      catch(SQLException sqx)
      {
        System.out.println("*** TLM got SQLException from getMoodleShortName()");
        System.out.println("*** moodleShortName set to nullstring");
        moodleShortName = "";
	      moodleFileName = "";
      }
    }// end slse/if (databasePrefix == null)
    
    // load downloadBuffer if there is a file matching MoodleShortName
    loadDownloadBuffer();
    
    // fill out host table to maxConnections
    highestReservedHost = numberOfHosts-1;
    for( ; numberOfHosts < maxConnections; numberOfHosts++)
    {
      initializeHostTableAt(numberOfHosts,false);
      anonymousHost[ numberOfHosts ] = !checkPassword;
      hostInetAddr[ numberOfHosts ] = null;
      hostAddr[ numberOfHosts ] = null;
      tunnelClosing[ numberOfHosts ] = false;
    }

    // setup authentication tables
    auth = new Auth();
    updateDBNumUsers(0,0); 

    // allocate mRecvTime array and zero it
    mRecvTime = new long[ numberOfApplications ];
    for( int entry = 0; entry < numberOfApplications; entry++ )
      mRecvTime[ entry ] = 0;

    // create sockets and instantiate threads to listen to them -
    // there will be two multicast sockets and two unicast UDP sockets
    // per application

    // none of the following is used when operating with stunnel
    if(!useSSL)
    {
      try
      {
        for( int applicationIndex = 0;
             applicationIndex < numberOfApplications && !stopTLM; 
             applicationIndex++)
        {
          // create multicast sockets with values parsed from file
          if( multicastIsConfigured && (mcPort[applicationIndex] > 1))
          {
            mcSocket1 = new MulticastSocket( mcPort[applicationIndex] );
            mcSocket1.setTimeToLive( 15 );
            mcSocket1Save[applicationIndex] = mcSocket1;
            mcSocket2 = new MulticastSocket( mcPort[applicationIndex]+1 );
            mcSocket2.setTimeToLive( 15 );
            mcSocket2Save[applicationIndex] = mcSocket2;
            mcSocket1.joinGroup( mcGroup[applicationIndex] );
            mcSocket2.joinGroup( mcGroup[applicationIndex] );
          }
           
          // create unicast sockets
          if( retransmitUDP ||
             (videoAddressConfigured && videoPort == udpPort[ applicationIndex ]) )
          {
            udpSocket1 = new DatagramSocket( udpPort[applicationIndex] );
            udpSocket1Save[applicationIndex] = udpSocket1;
            udpSocket2 = new DatagramSocket( udpPort[applicationIndex]+1 );
            udpSocket2Save[applicationIndex] = udpSocket2;
          }
          // instantiate two MCListen threads per application
          if( multicastIsConfigured && (mcPort[applicationIndex] > 1))
          {
            MCListen s1 = new MCListen(applicationIndex,
                                       udpPort[applicationIndex],
                                       mcPort[applicationIndex],
                                       mcSocket1,
                                       udpSocket1 );
            MCListen s2 = new MCListen(applicationIndex,
                                       udpPort[applicationIndex]+1,
                                       mcPort[applicationIndex]+1,
                                       mcSocket2,
                                       udpSocket2 );
            Thread t1 = new Thread(s1);
            Thread t2 = new Thread(s2);

            // reduce priority of multicast threads, except the lowest one (audio)
            // so they will be interrupted by the FCListen, PLAYBACK and UCListen
            // threads, all of which have a way to sleep while waiting on data
            // (multicast does not)
            int mPriority = t1.getPriority();
            --mPriority;
            if( applicationIndex > 0 )
              t1.setPriority( mPriority );
            t2.setPriority( mPriority );
            t1.start();
            t2.start();
          }
          // if retransmitting UDP, instantiate two UCListen threads per application
          if(retransmitUDP ||
             (videoAddressConfigured && (videoPort == udpPort[ applicationIndex ])) )
          {
            UCListen r3 = new UCListen(applicationIndex,
                                       udpPort[applicationIndex],
                                       mcPort[applicationIndex],
                                       mcSocket1,
                                       udpSocket1 );
            UCListen r4 = new UCListen(applicationIndex,
                                       udpPort[applicationIndex]+1,
                                       mcPort[applicationIndex],
                                       mcSocket2,
                                       udpSocket2 );
            Thread t3 = new Thread(r3);
            Thread t4 = new Thread(r4);
            // increase priority of the first (audio) application (but not control)
            int uPriority = t3.getPriority();
            if( applicationIndex == 0 )
              t3.setPriority( ++uPriority );
            t3.start();
            t4.start();
          } // end if( retransDP )
        } // end for( applicationIndex
        
        // multicast tunnel socket with listener thread
        if(multicastTunnelServerSubnet)
        {
          mcTunnelRecvSocket = new MulticastSocket(multicastTunnelPort);
          mcTunnelRecvSocket.joinGroup(multicastTunnelGroup);
          mcTunnelSendSocket = new MulticastSocket();
          
          // disable loopback on multicast send
          mcTunnelSendSocket.setLoopbackMode(true);
          if(mcTunnelSendSocket.getLoopbackMode())
          {
            mctl = new MCTunnelListen();
            Thread mcTunnelThread = new Thread(mctl);
            mcTunnelThread.setPriority(mcTunnelThread.getPriority()-1);
            mcTunnelThread.start();
          }
          else
            System.out.println("*** unable to disable multicast loopback\n" +
              " server will send but not receive tunnel multicast");
        }
      } // end try for setting up sockets
      catch( SocketException e )
      {
        System.out.print( "Socket creation error: " + e.getMessage() + "\n" );
        System.out.print( "Execution aborted.\n" );
        stopTLM = true;
      }
      catch( IOException e )
      {
        // error message received when Unicast port is not open and execution halts
        System.out.print( "Port not open error: " + e.getMessage() );
        stopTLM = true;
      }
      // done creating UDP ports & thread
    } // end if(!useSSL)

    // if using multicast floor control, start a thread for that
    Thread t3;
    if(mcFcInetAddr != null)
    {
       mcfcListen = new MCFCListen();
///* irmp
       // instantiate the Improved Reliability Multicast Protocol
       irmp = new IRMP(mcFcPort,
                       mcFcInetAddr);
       t3 = new Thread(irmp);
       t3.start();
//irmp */
/* irmp
       t3 = new Thread(mcfcListen);
       t3.start();
irmp */
       usingMcFC = true;
    }
    if(stopTLM)System.exit(1);
    int hostIndex = 0;

    // arrange to catch shutdown command
    Runtime.getRuntime().addShutdownHook
    ( new Thread()
      {
        public void run()
        {
          try
          {
            TLM.close();
          }
          catch( Exception e ){}
        } // end run()
      } // end thread
    ); // end Shutdown Hook

    try
    {
      // obtain a SocketChannel on tunnelTCPPort
      new NEWSocketChannel( tunnelTcpPort, maxConnections, useSSL);
    }
    catch( Exception e )
    {
      System.out.println( "Unable to create server sockets in TLM, port:" +
                          tunnelTcpPort + " exception:" + e);
      stopTLM = true;
      System.exit(1);
    }

    // spawn FCConnect thread to accept incoming floor control connections
    if( !stopTLM )
    {
      fcConnect = new FCConnect();
      Thread t0 = new Thread(fcConnect);
      //t0.start();

      // spawn Heartbeat thread so we always have some messages coming
      // (needed so the TCP loop waiting on TCPread.available can get interrupted)
      Heartbeat h1 = new Heartbeat();
      Thread th = new Thread(h1);
      th.start();

      if(!useSSL)
      {
        // instantiate the UDPTunnel thread to listen to TLM_UDP_Port
        UDPTunnel u1 = new UDPTunnel();
        Thread tu = new Thread(u1);
        tu.start();

        try
        {
          // create a port to use in floor control warnings through UDP tunnels
          tunnelMessageUDPSocket = new DatagramSocket();
        }
        catch( IOException ioe )
        {
          System.out.println( "Unable to create FC UDP socket in TLM");
          stopTLM = true;
          System.exit(1);
        }
      }

      // if running auto scheduler, start a thread to time server 
	    // operation; also start recorder if configured
      if(autoSchedule)
      {
        CourseTimer courseTimer = new CourseTimer();
        Thread classTimerThread = new Thread(courseTimer);
        classTimerThread.start();
      }
      else
		  // without autoschedule, default courseID to roomID
		  thisCourseID = roomID;

      // print startup to log
      System.out.print( "\nTunnel server waiting for client on TCP port " +
        NEWSocketChannel.getDataSocketChannelLocalPort() );
      if( (tunnelUdpPort != 0) && !useSSL )
      System.out.print( " and UDP port " + tunnelUdpPort );
      System.out.print( "\n" );
      if( videoPort != 0 )
        System.out.println( "also accepting second TCP connection for video" );
      if( videoAddressConfigured )
        System.out.println( "also accepting a single video input from "
          + videoInetAddress + "/" + videoPort );
      System.out.println();

      // accept connections in an infinite loop
      while ( !stopTLM )
      {
        SocketChannel acceptSocketChannel;
        try
        {
          // open a dedicated socket for next connecting host
          acceptSocketChannel = NEWSocketChannel.getDataSocketChannel();
          if(acceptSocketChannel == null)continue;
        }
        catch( Exception e )
        {
          // lost connection while accepting
          continue;
        }
        // don't start a new connection while closeAllConnections is running
        synchronized(recvSocketLock)
        {
          tcpRecvSocketChannel = acceptSocketChannel;
        }
        // start a thread to deal with this connection
        TCPHandoff handoff = new TCPHandoff(tcpRecvSocketChannel);
        Thread handoffThread = new Thread(handoff);
        handoffThread.start();
        tcpConnectionsMade++;
      } // end while( !stopTLM )
    } // end if( !stopTLM )
  }// end main


  /*************************** downloadBuffer ***********************************/
  // read in the appFileDownload if there is one
  // returns true on failure, false otherwise
  static boolean loadDownloadBuffer()
  {
    int hostIndex;
    
    // if not configured for download, don't
    if(appFileDownloadLimit <= 0)return true;

    // buffering is not interruptable
    if(downloadBuffering)return true;
    downloadBuffering = true;

    // download is disabled until we establish a usable file size
    downloadIsReady = false;
    appFileDownloadSize = 0;
    
    // break off any downloads underway
    for(hostIndex = 0; hostIndex < maxConnections; hostIndex++)
    {
      if(tunnelSends[hostIndex] != null)
      {
        tunnelSends[hostIndex].proceedWithDownload = false;
        try
        {
          tunnelSends[hostIndex].sendDownloadDone(true);
        }
        catch(IOException ioe){}
      }
    }

    // we need a courseID to find the download file
    if(moodleFileName == null)
    {
      System.out.println("\n*** moodleShortName not set so can't load download file");
      downloadBuffering = false;
      return true;
    }    
  	String appFilePathName =
  	  new String(appFilePath + moodleFileName + "/" + moodleFileName + ".zip");
    
  	// confirm file exists
  	File appZipFile = new File(appFilePathName);
  	if(!appZipFile.exists())
  	{
	    System.out.println("\n*** application download zipfile " + appFilePathName +
	                       " not accessible ");
	    downloadBuffering = false;
            return true;
    }
    
    // obtain file parameters
    appFileDownloadName = moodleFileName+".zip";
  	long appZipFileLength = appZipFile.length();
  	appFileDownloadMtime = appZipFile.lastModified()/1000;
  	appFileDownloadMtimeBytes = intToBytes((int)(appFileDownloadMtime));
  	if(appZipFileLength > appFileDownloadLimit)
  	{
      System.out.println("\napplication download zipfile " + appFilePathName +
                         " + length " + appZipFileLength + " exceeds configured max "
                         + appFileDownloadLimit);
      downloadBuffering = false;
      return true;
    }
    appFileDownloadLengthBytes = intToBytes((int)(appZipFileLength));
    
    // open a stream on the file
    if(appFileDownload == null)
  	  appFileDownload = new byte[(int)appFileDownloadLimit];
  	FileInputStream appZipStream = null;
  	try
  	{
      appZipStream = new FileInputStream(appZipFile);
    }
    catch(FileNotFoundException fnfe)
    {
 	    System.out.println("\n*** application download zipfile " + appFilePathName +
                         " not found");
      downloadBuffering = false;
      return true;
    }
    // read in the file
    try
    {
      for (long byteNumber = 0;
           byteNumber < appZipFileLength;
           byteNumber += DOWNLOAD_MESSAGE_SIZE)
      {
        long bytesToRead = appZipFileLength - appFileDownloadSize;
        if(bytesToRead > DOWNLOAD_MESSAGE_SIZE)bytesToRead = DOWNLOAD_MESSAGE_SIZE;
        long bytesRead = appZipStream.read(appFileDownload,
                                           (int)byteNumber,(int)bytesToRead);
        if(bytesRead < 0)break;
        appFileDownloadSize += bytesRead;
	  }
    }
    catch(IOException ioe)
    {
      System.out.println("\n*** IOException reading application download zipfile \"" +
  	                     appFilePathName + "\" :" + ioe);
      appFileDownloadSize = 0;
      downloadBuffering = false;
      return true;
    }
    
    // tell TunnelSend the download is ready
    if(appFileDownloadSize == 0)
    {
      downloadBuffering = false;
      return false;
    }
    downloadIsReady = true;
    System.out.println("\ncached application data zipfile \"" + appFilePathName +
                       "\" length " + appFileDownloadSize + " bytes" +
					             " mtime " + appFileDownloadMtime);
    
    // start downloads in active TunnelSend instances
    for(hostIndex = 0; hostIndex < maxConnections; hostIndex++)
    {
      try
      {
        if(tunnelSends[hostIndex] != null)
          tunnelSends[hostIndex].startDownloading();
      }
      catch(Exception e)
      {
        System.out.println("*** failed to start download for host:" +
                           hostIndex + " exception:" + e);
        return true;
      }
	  }// end for(hostIndex = 0;

    // set buffering status off      
    downloadBuffering = false;
    
    return false;

  } // end downloadBuffer
  
  
  // load the download file into buffer
  class ReBuffer extends Thread
  {
	  public void run()
	  {
      // don't run loadDownloadBuffer when it's already running
      try
      {
        while(downloadBuffering)
          Thread.sleep(500);
      }
      catch(InterruptedException ie){return;}
      loadDownloadBuffer();
    }
  }// end ReBuffer


  /************************* closeAllConnections ******************************/
  // shutdown active connections so server can stop or take on a new class
  static void closeAllConnections(String reason)
  {
    // don't close any connections while a new connection is being accepted
    synchronized (recvSocketLock)
    {
      // close any open tunnel
      for (int hostIndex = 0; hostIndex < numberOfHosts; ++hostIndex)
        if( tcpIsOpen[hostIndex] || (hostInetAddr[hostIndex] != null) )
          closeTunnel(hostIndex, false, reason);
	    
      // reset the hashtable of users
      FCConnect.clearActiveUsers();
       
      // reset download status
      downloadOfferedToController = false;
      
      // reset host table
      numberOfHosts = highestReservedHost+1;
      for( ; numberOfHosts < maxConnections; numberOfHosts++)
      {
        initializeHostTableAt(numberOfHosts,false);
        anonymousHost[numberOfHosts] = !checkPassword;
        hostInetAddr[numberOfHosts] = null;
        hostAddr[numberOfHosts] = null;
        tunnelClosing[numberOfHosts] = false;
        username[ numberOfHosts ] = "        ";
        if (numberOfHosts > highestReservedHost) 
        {
          controlPrivilege[numberOfHosts] = '0';
          instructorPrivilege[numberOfHosts] = 0;
          drawGrid[numberOfHosts] = false;
        }
        appFileMtime[numberOfHosts] = 0;
        appFileSize[numberOfHosts] = 0;
        appFileStatNew[numberOfHosts] = false;
        breakoutGroup[numberOfHosts] = 0;
        netCapacitykbps[numberOfHosts] = 0;
      }

      // zero tunnel/flow counts
      setUserCounts(0,0);
      udpTunnelsMade = 0;
      udpFlowsActive = 0;
	  
	    // reset breakout
	    breakoutController = -1;
	    instructorBreakoutGroup = 0;
      breakoutDataMessage = null;
          
    } // end synchronized (recvSocketLock)
    
  } // end closeAllConnections()
  
  
  /****************************** close **************************************/
  //  termination processing- close sockets and shut down by setting "stopTLM"
  static void close() throws IOException
  {
    timeEnd = System.currentTimeMillis();
    System.out.print("Shutdown command received- TLM closing connections and exiting.\n");
    closeAllConnections("shutdown");

    updateDBNumUsers(0,0); 

    // close sockets and interrupt any thread that might be sleeping
    try
    {
      if( tcpRecvSocketChannel != null )tcpRecvSocketChannel.close();
      for( int appNumber = 0; appNumber < numberOfApplications; ++appNumber)
      {
        if( mcSocket1Save[ appNumber ] != null )mcSocket1Save[ appNumber ].close();
        if( mcSocket2Save[ appNumber ] != null )mcSocket2Save[ appNumber ].close();
        if( udpSocket1Save[ appNumber ] != null )
          udpSocket1Save[ appNumber ].close();
        if ( udpSocket2Save[ appNumber ] != null )
          udpSocket2Save[ appNumber ].close();
      }
      for( int hostIndex = 0; hostIndex < numberOfHosts; ++hostIndex)
        if(tcpSocketChannels[ hostIndex ] != null)
          tcpSocketChannels[ hostIndex ].close();
    }
    catch( IOException e )
    {
      // lost connection - already shutting down - do nothing
    }
    // calculate and print performance
    System.out.print("\n\nTLM statistics for server operation "
      + startupDateTime + " to " + dateTime() + ":\n");
    float deltaR = 1000/(float)(timeEnd - timeStart);
    System.out.print("Multicast messages received:" + countMcRecv + " pps: "
      + ((float)countMcRecv*deltaR) + "\tpackets/sec\n");
    System.out.print("Multicast messages received:" + countMcSend + " pps: "
      + ((float)countMcSend*deltaR) + "\tpackets/sec\n");
    System.out.print("UDP messages received:" + countUdpRecv + " pps: "
      +((float)countUdpRecv*deltaR) +"\tpackets/sec\n");
    System.out.print("UDP messages sent:" + countUdpSend + " pps: "
      +((float)countUdpSend*deltaR) +"\tpackets/sec\n");
    System.out.print("UDP in UDP tunnel messages received:" + countUdpTunnelRecv + " pps: "
      +((float)countUdpTunnelRecv*deltaR) +"\tpackets/sec\n");
    System.out.print("UDP in UDP tunnel messages sent:" + countUdpTunnelSend + " pps: "
      +((float)countUdpTunnelSend*deltaR) +"\tpackets/sec\n");
    System.out.print("UDP in TCP tunnel messages received:" + countTcpRecv + " pps: "
      +((float)countTcpRecv*deltaR) +"\tpackets/sec\n");
    System.out.print("UDP in TCP tunnel messages sent:" + countTcpSend + " pps: "
      +((float)countTcpSend*deltaR) +"\tpackets/sec\n");
    System.out.print("Total TCP tunnel connections: " + tcpConnectionsMade + "\n");
    System.out.print("Total UDP tunnel connections: " + udpTunnelsMade + "\n");
    System.out.println("TLM shut down at:" + dateTime());

    // signal all other parts of server to quit
    stopTLM = true;

  } // end TLM/close() method


  /*************************** closeTunnel ***********************************/
  // close connection to specified client
  //
  // closing is syncrhonized by tunnelClosing[clientIndex]
  // use of the tunnel itself stops when tunnelIsActive[clientIndex] is false
  // user state is retained until initializeAutoReconnect()
  // host state is reatined until clearHostDataFromHostTable() 
  static Object tunnelClosingSynch = new Object();
  static void closeTunnel(int clientIndex, boolean unexpected, String reason)
  {
    String protocol;
          
	  // the following should be reinitialized by exception in any  JMP 13 Dec 09
	  // case where auto-reconnection will not be attempted:
	  // authenticatedHost[clientIndex] = false;
	  // tcpSocketChannels[clientIndex] = null;
    // sessionVariable[clientIndex] = "";
	  //
  	// reinitialize these after printling log message:
	  // controlPrivilege[clientIndex] = "";
	  // instructorPrivilege[clientIndex] = 0;
    // drawGrid[clientIndex] = false;
	  // synchronized(hostInetAddr)
    // {
    //   hostInetAddr[clientIndex] = null;
    //   hostAddr[clientIndex] = null;
    // }
	
    // only need to close it once
    if(clientIndex < 0)
	  {
	    System.out.println("*** closing tunnel with invalid index:"+clientIndex+
                         " message:"+reason);
      return;
    }
	
    // lock use of tunnel status arrays to avoid conflicts among
	  // multiple closes; then flag this tunnel inactive
    tunnelIsActive[clientIndex] = false;
    synchronized(tunnelClosingSynch)
    {
      if(tunnelClosing[clientIndex])return;
      tunnelClosing[clientIndex] = true;
    }
	
    // shut down associated TCP tunnel if any   
    if(tcpTunnelThreads[clientIndex] != null)
    {
      try
      {
        if(tunnelSends[clientIndex] != null)
          tunnelSends[clientIndex].sendBuffer(shutdown,4,false);
      }catch(IOException ioe){}
      tcpTunnelThreads[clientIndex].interrupt();
      tcpTunnelThreads[clientIndex] = null;
      try{tcpSocketChannels[clientIndex].close();}catch(IOException ioe){}
      decrementUserCounts(clientIndex);
      
	  }
    tcpIsOpen[clientIndex] = false;
    
    // and TCP video tunnel
    if(videoThreads[clientIndex] != null)
    {
      videoThreads[clientIndex].interrupt();
      videoThreads[clientIndex] = null;
      try 
      {
        tcpVideoSocketChannels[clientIndex].close();
      }
      catch(IOException ioe) {}
    }
    
	  // if unexpected disconnect and this host had floor, 
	  // save its index and give floor to nobody
	  if(unexpected)
	    if(floorHolder == clientIndex)
	      previousFloorHolder = clientIndex;
	
	  // notify associated TunnelSend to shut down
	  if(sendThreads[clientIndex] != null)
	    synchronized (sendThreads[clientIndex])
	    {
        sendThreads[clientIndex].notify();
		    tunnelSends[clientIndex] = null;
      }
      
    // reset Breakout mode if this was breakoutController
    if(clientIndex == breakoutController)
      fcConnect.resetBreakoutMode(clientIndex);
    	
	  // delete vote closing state if disconnected
	  if(openVoteFloorholder == clientIndex)
	  {
	    openVoteFloorholder = -1;
	    openVoteDialog = null;
	    startAVote = null;
	    fcConnect.writeMessageToAll(21,0,null,false);
    }
    try
    {
      // tell all floor controls this one is gone
      fcConnect.dropParticipant(clientIndex);
      
      // update floorHolder status
      if(clientIndex == floorHolder)
        if(inBreakoutMode)
          fcConnect.setFloorHolder(-1,"********",clientIndex,breakoutGroup[clientIndex]);
        else
          fcConnect.setFloorHolder(-1,"********",clientIndex,-1);            
	  
      // capture current state of this tunnel
      String closingHost = hostAddr[clientIndex];
      if (protocolIsTCP[clientIndex])
      {
        if (connectedVideoByTCP[clientIndex])
          protocol = "VTCP";
        else
          protocol = "TCP";
      }
      else
      {
        protocol = "UDP";

        // send shutdown message to TLMC
        if(hostInetAddr[clientIndex] != null)
        {
          byte[] heartbeatBuffer = {68, 0, 0, 1, 0};
          DatagramPacket sendMessage =
              new DatagramPacket(heartbeatBuffer, 5,
                                 hostInetAddr[clientIndex],
                                 tunnelUdpPort + 1);
          try
          {
            if (udpSockets[clientIndex] != null)
              udpSockets[clientIndex].send(sendMessage);
          }
          catch (IOException ioe) {}
          countUdpSend++;
        }
      }

      // quit using the connection
      // find position this connection occupies in tcpActive array
      // shut off semaphores and send thread
      
      // the entry might not be there if it never authenticated
      if( connectedByTCP[clientIndex] )
      {
        // turn off flag and check counter (bugproofing)
        connectedByTCP[ clientIndex ] = false;
        if (numberOfTunnelsActive < 0)
        {
          System.out.println("*** numberOfTunnelsActive negative on host #"
                             + clientIndex);
          setUserCounts(0,0);
          
        } // end if( numberOfTunnelsActive...
      } // end if(!connectedByTCP...
      else if( !udpActive[clientIndex] )
             deactivateFlow(clientIndex);

      // set no-traffic condition
      uRecvTime[clientIndex] = 0;

      // also shut down floor control connection if possible
      if (fcOutStreams[clientIndex] != null)
      {
        try
        {
          fcOutStreams[clientIndex].close();
        }
        catch (IOException e) {}
        fcOutStreams[clientIndex] = null;
      }
      if (fcInStreams[clientIndex] != null)
      {
        try
        {
          fcInStreams[clientIndex].close();
        }
        catch (IOException e) {}
        fcInStreams[clientIndex] = null;
      }
      // reset sockets for the connection
      try
      {
        if (fcSockets[clientIndex] != null)
          fcSockets[clientIndex].close();
      }
      catch (IOException e) {}
      fcSockets[clientIndex] = null;
      try
      {
        if (tcpSocketChannels[clientIndex] != null)
          tcpSocketChannels[clientIndex].close();
      }
      catch (IOException e) {}    
      try
      {
        if (tcpVideoSocketChannels[clientIndex] != null)
          tcpVideoSocketChannels[clientIndex].close();
      }
      catch (IOException ioe) {}
      udpSockets[clientIndex] = null;
      tcpVideoSocketChannels[clientIndex] = null;
      connectedVideoByTCP[clientIndex] = false;
      otherEndPort[clientIndex] = -1;
      setupReceived[clientIndex] = false;

      // empty the ring buffers for this host 
      int bufferIndex;
	    for (bufferIndex = 0;
		    bufferIndex < audioRingSize;
		    ++bufferIndex)
	    {
		    audioQueues[clientIndex][bufferIndex] = null;
		    audioLengths[clientIndex][bufferIndex] = 0;
	    }
	    for (bufferIndex = 0;
		    bufferIndex < fcRingSize;
		    ++bufferIndex)
	    {
		    fcQueues[clientIndex][bufferIndex] = null;
		    fcLengths[clientIndex][bufferIndex] = 0;
	    }
      for (bufferIndex = 0;
           bufferIndex < whiteboardRingSize;
           ++bufferIndex)
      {
		    whiteboardQueues[clientIndex][bufferIndex] = null;
        whiteboardLengths[clientIndex][bufferIndex] = 0;
	    }
      for (bufferIndex = 0;
		       bufferIndex < videoRingSize;
		       ++bufferIndex)
      {
		    videoQueues[clientIndex][bufferIndex] = null;
		    videoLengths[clientIndex][bufferIndex] = 0;
      }
      for (bufferIndex = 0;
           bufferIndex < multicastRingSize;
           ++bufferIndex)
      {
		    multicastQueues[clientIndex][bufferIndex] = null;
		    multicastLengths[clientIndex][bufferIndex] = 0;
      }
      
      // write disconnect to log
      long elapsedTime = (System.currentTimeMillis() - loginTime[clientIndex]) /
          1000;
      if(unexpected)
        System.out.print("\n" + "NEWLIVE3:" + dateTime() + " user:"
                         + username[clientIndex] + " on #" + clientIndex
                         + " host:" + closingHost + " " + protocol
                         + " disconnected after duration:" + elapsedTime
                         + " seconds PeakRTT:" + peakRTT[clientIndex]
                         + " remaining TCP connected users: "
                         + numberOfTunnelsActive
                         + " number of controllers remaining:"
                         + numberOfControllersLoggedIn
                         + " reason:" + reason + "\n");
      else
        System.out.print("\n" + "NEWLIVE2:" + dateTime() + " user:"
                         + username[clientIndex] + " on #" + clientIndex
                         + " host:" + closingHost + " " + protocol
                         + " disconnected after duration:" + elapsedTime
                         + " seconds PeakRTT:" + peakRTT[clientIndex]
                         + " remaining TCP connected users: "
                         + numberOfTunnelsActive 
                         + " number of controllers remaining:"
                         + numberOfControllersLoggedIn
                         + " reason:" + reason + "\n");

      // write to database log if there has been an authenticated connection
      if (!username[clientIndex].equals("        ")) 
      {
	      updateDBNumUsers(numberOfTunnelsActive,numberOfControllersLoggedIn);
        addToDatabase("2", elapsedTime, username[clientIndex],
                      closingHost, userCourseID[clientIndex],
                      controlPrivilege[clientIndex], protocol,
                      peakRTT[clientIndex]);
      }      		

      // clean up status arrays
      if(!authenticatedHost[clientIndex])
        clearHostDataFromHostTable(clientIndex);
      if(reason.length() >= 22)
        if(reason.substring(0,22).equals("client sent disconnect"))
        {
          clearUserDataFromHostTable(clientIndex);
          clearHostDataFromHostTable(clientIndex);
        }
      initializeHostTableAt(clientIndex,true);

    }
    catch(ArrayIndexOutOfBoundsException ae)
    {
      System.out.println("array index out of bounds in closeTunnel for host " +
                         clientIndex + "\nreason:" + reason);
      ae.printStackTrace(System.out);
    }
    catch(Exception e)
    {
      System.out.println("Exception closing tunnel " + clientIndex + ":" + e +
                         "\nreason:" + reason);
      e.printStackTrace(System.out);
    }
    
    // unlock tunnelClosing
    synchronized( tunnelClosingSynch )
    {
      tunnelClosing[ clientIndex ] = false;
    }
    
    // start autoReconnect timer
    if(tcpHandoffs[clientIndex]  != null)
      tcpHandoffs[clientIndex].clearThisConnectionAfterWait(clientIndex);
   
} // end closeTunnel method


  /*************************** activateTunnel ********************************/
  // when a new TCP or UDP tunnel has been validated
  static synchronized void activateTunnel(int clientIndex)
  {
    TunnelSend s = new TunnelSend( clientIndex );
    tunnelSends[clientIndex] = s;
    Thread t = new Thread(s);
    sendThreads[ clientIndex ] = t;
    t.start();

    // this method can be called either from TCPTunnel or UDPTunnel;
    // if from TCPTunnel, tcpIsOpen will be true
    if(tcpIsOpen[clientIndex])incrementUserCounts(clientIndex);
    else udpTunnelsMade++;
    
    // write new connection to log in parsable format
    String protocol = protocolIsTCP[ clientIndex ]?"TCP":"UDP";
    String printController = "";
    if( controlPrivilege[ clientIndex ] > '0' )
      printController = " ** floor controller **";
    System.out.println( "\n" + "NEWLIVE1:" + dateTime() + " user:" +
                        username[ clientIndex ] + " accepted by " + protocol +
                        " on #" + clientIndex + " host:" + hostAddr[ clientIndex ] +
                        "/" + otherEndPort[ clientIndex ] + printController);
    System.out.println("total TCP connected users: " + numberOfTunnelsActive +
                       " total controllers:" + numberOfControllersLoggedIn);
    peakRTT[ clientIndex ] = 0;
    
    // keep track of number of users with floor control privileges
    if(fcConnect.getControlPrivilege(clientIndex) > '0')
      if(!nameExistsAgain[clientIndex])
        numberOfControllersLoggedIn++;

    // write to database log
    addToDatabase("1", 0, username[ clientIndex ],
                  hostAddr[ clientIndex ],
                  userCourseID[ clientIndex ],
                  controlPrivilege[ clientIndex ],
                  protocol, 0 );

  } // end activateTunnel method


  /**************************** activateFlow **********************************/
  // when a new UDP flow has been validated
  static synchronized void activateFlow( int clientIndex,
                                         InetAddress clientAddress )
  {
    if( clientAddress == null )return; // host disconnected during startup
    udpActive[ udpFlowsActive++ ] = true;
    if( anonymousHost[ clientIndex ] && (clientIndex > highestReservedHost) )
    {
      System.out.println("Started multicasting anonymous input from "
        + hostAddr[ clientIndex ] + " as host# " + clientIndex);
    }
  } // end activateFlow method

  // method invoked when a UDP flow stops sending
  static synchronized void deactivateFlow(int clientIndex)
  {
    udpActive[ clientIndex ] = false;

    // if this is an anonymous position, reset the hostAddr value
    if( anonymousHost[ clientIndex ] && (clientIndex > highestReservedHost) )
    {
      System.out.println("stopped forwarding anonymous input from "
        + hostAddr[ clientIndex ]+ " as host# " + clientIndex);
      synchronized(hostInetAddr)
      {
        hostInetAddr[clientIndex] = null;
      }
      synchronized(hostAddr)
      {
        hostAddr[clientIndex] = null;
      }
    }

    // decrement flows active and return
    udpFlowsActive--;

  } // end deactivateFlow method


  /***************************** queueTunnel *******************************/
  // method to enqueue a buffer if the host is a legitimate receiver
  // returns true if no space for the buffer in queue array
  static synchronized boolean queueTunnel( int priority, int thisHostIndex,
                                           byte[] sendBuffer, int length )
  {
    int top, bottom, holdTop = 0;
    if( !tunnelIsActive[ thisHostIndex ] ) return false;
    if( priority == 1 )
    {
      // check for full audio window
      synchronized(audioQueues[thisHostIndex])
      {
        bottom = audioWindowStart[ thisHostIndex ];
        top = audioWindowEnd[ thisHostIndex ];
        holdTop = top;
        if( top < bottom ) top = top + audioRingSize;
        if( (top - bottom) >= (audioRingSize - 1) ) return true;

        // save this buffer and its length in highest (audio) priority
        audioQueues[thisHostIndex][holdTop] = sendBuffer;
        audioLengths[thisHostIndex][holdTop] = length;
        if (++holdTop >= audioRingSize)
          holdTop = 0;
        audioWindowEnd[thisHostIndex] = holdTop;
        
      }// end synchronized
    }
    else if( priority == 2 )
    {
      // check for full floor control window
      synchronized (fcQueues[thisHostIndex])
      {
        bottom = fcWindowStart[thisHostIndex];
        top = fcWindowEnd[thisHostIndex];
        holdTop = top;
        if (top < bottom)
          top = top + fcRingSize;
        if ( (top - bottom) >= (fcRingSize - 1))
          return true;

        // save this buffer and its length in middle (whiteboard) priority
        fcQueues[thisHostIndex][holdTop] = sendBuffer;
        fcLengths[thisHostIndex][holdTop] = length;
        if (++holdTop >= fcRingSize)
          holdTop = 0;
        fcWindowEnd[thisHostIndex] = holdTop;
        
      } // end synchronized
    }
    else if( priority == 3 )
    {
      // check for full whiteboard window
      synchronized (whiteboardQueues[thisHostIndex])
      {
        bottom = whiteboardWindowStart[thisHostIndex];
        top = whiteboardWindowEnd[thisHostIndex];
        holdTop = top;
        if (top < bottom)
          top = top + whiteboardRingSize;
        if ( (top - bottom) >= (whiteboardRingSize - 1))
          return true;

        // save this buffer and its length in middle (whiteboard) priority
        whiteboardQueues[thisHostIndex][holdTop] = sendBuffer;
        whiteboardLengths[thisHostIndex][holdTop] = length;
        if (++holdTop >= whiteboardRingSize)
          holdTop = 0;
        whiteboardWindowEnd[thisHostIndex] = holdTop;
        
      } // end synchronized
    }
    else if(priority == 4)
    {
      // check for full video window
      synchronized(videoQueues[thisHostIndex])
      {
        bottom = videoWindowStart[thisHostIndex];
        top = videoWindowEnd[thisHostIndex];
        holdTop = top;
        if (top < bottom)
          top = top + videoRingSize;
        if ( (top - bottom) >= (videoRingSize - 1))
          return true;
         
        // save this buffer and its length in bottom (video) priority
        videoQueues[thisHostIndex][holdTop] = sendBuffer;
        videoLengths[thisHostIndex][holdTop] = length;
        if (++holdTop >= videoRingSize)
          holdTop = 0;
        videoWindowEnd[thisHostIndex] = holdTop;
        
      }// end synchronized
    }
    else
    {
      // check for full multicast window
      synchronized(multicastQueues[thisHostIndex])
      {
        bottom = multicastWindowStart[thisHostIndex];
        top = multicastWindowEnd[thisHostIndex];
        holdTop = top;
        if (top < bottom)
          top = top + multicastRingSize;
        if ( (top - bottom) >= (multicastRingSize - 1))
          return true;

        // save this buffer and its length in sub-bottom (multicast) priority
        multicastQueues[thisHostIndex][holdTop] = sendBuffer;
        multicastLengths[thisHostIndex][holdTop] = length;
        if (++holdTop >= multicastRingSize)
          holdTop = 0;
        multicastWindowEnd[thisHostIndex] = holdTop;
        
      }// end synchronized
    }

    // if this is the only packet in the queue, wake up the send method
    if( top == bottom )
    {
      if (sendThreads[ thisHostIndex ] != null)
      {
        synchronized (sendThreads[thisHostIndex])
        {
          sendThreads[thisHostIndex].notify();
        }
      }
    }
    return false;

  } // end queueTunnel method


  /**************************** sendFloorWarning ***************************/
  // method to send a floor warning message
  // returns true if exception during sending, false otherwise
  boolean sendFloorWarning( int hostIndex, int portNum )
  {
    if(!tcpIsOpen[ hostIndex ] )return false;
    // message to warn sending without being floor holder
    byte[] errorBuffer = {69,0,0,2,0};
    
    // print to log and send message
    System.out.println("floor warning to host #" + hostIndex
      + " due to traffic on encapsulated port " + portNum );
    try
    {
      if( tcpIsOpen[ hostIndex ] )
      {
        queueTunnel(2,hostIndex,errorBuffer,5);
      }
      else if(hostInetAddr[ hostIndex] != null)
      {
        DatagramPacket errorMessage =
          new DatagramPacket( errorBuffer, 5,
                              hostInetAddr[ hostIndex ],
                              tunnelUdpPort+1 );
        tunnelMessageUDPSocket.send( errorMessage );
      }
    }
    catch( Exception e )
    {
      // lost connection
      closeTunnel( hostIndex, true, "floor warning general exception" );
      return true;
    }
    return false;
  } // end sendFloorWarning method


  /************************** host table utilities *************************/
  
  // initialize variables in host table that normally are not changed
  // by initializeHostTableAt because they're needed for autoReconnect
  static void initializeAutoReconnect(int hostIndex)
  {
    tcpSocketChannels[ hostIndex ] = null;
    username[ hostIndex ] = "        ";
    fullname[ hostIndex ] = "";
    sessionVariable[ hostIndex ] = "";
    authenticatedHost[ hostIndex ] = false;
    tcpHandoffs[ hostIndex ] = null;
    tcpTunnelInstance[ hostIndex ] = null;
    breakoutGroup[ hostIndex ] = -1;
    fcIsInitialized[ hostIndex ] = false;
    fcIsReadyForAppData[ hostIndex ] = false;
    controlPrivilege[ hostIndex ] = 0;
    connectionTestPassed[ hostIndex ] = false;

  }// end initializeAutoReconnect()

  // initialize a host table position
  static void initializeHostTableAt(int hostIndex,boolean autoReconnect)
  {
    if(!autoReconnect)
	  initializeAutoReconnect(hostIndex);
    setupReceived[ hostIndex ] = false;
    latestRTTIncrease[ hostIndex ] = 0;
    audioDropped[ hostIndex ] = 0;
    whiteboardDropped[ hostIndex ] = 0;
    lastErrorMessageType[ hostIndex ] = -1;
    videoDropped[ hostIndex ] = 0;
    multicastDropped[ hostIndex ] = 0;
    udpSockets[ hostIndex ] = null;
    tcpVideoSocketChannels[ hostIndex ] = null;
    fcSockets[ hostIndex ] = null;
    fcInStreams[ hostIndex ] = null;
    fcOutStreams[ hostIndex ] = null;
    sendThreads[ hostIndex ] = null;
    tunnelSends[ hostIndex ] = null;
    sourcePort[ hostIndex ] = 0;
    sourceVideoPort[ hostIndex ] = 0;  
    otherEndPort[ hostIndex ] = 0;
    userCourseID[ hostIndex ] = "           ";
    loginTime[ hostIndex ] = System.currentTimeMillis();
    killConnection[ hostIndex ] = false;
    tunnelIsActive[ hostIndex ] = false;
    tcpIsOpen[ hostIndex ] = false;
    connectedByTCP[ hostIndex ] = false;
    connectedVideoByTCP[ hostIndex ] = false;
    udpSetupTime[ hostIndex ] = 0;
    udpLoopbackCount[ hostIndex ] = 0;
    uRecvTime[ hostIndex ] = 0;
    fRecvTime[ hostIndex ] = -1;
    fcIsMc[ hostIndex ] = false;
    peakRTT[ hostIndex ] = 0;
    lastWarningTime[ hostIndex ] = 0;
    audioWindowStart[ hostIndex ] = 0;
    audioWindowEnd[ hostIndex ] = 0;
    fcWindowStart[ hostIndex ] = 0;
    fcWindowEnd[ hostIndex ] = 0;
    videoWindowStart[ hostIndex ] = 0;
    videoWindowEnd[ hostIndex ] = 0;
    whiteboardWindowStart[ hostIndex ] = 0;
    whiteboardWindowEnd[ hostIndex ] = 0;
    multicastWindowStart[ hostIndex ] = 0;
    multicastWindowEnd[ hostIndex ] = 0;
    numAuthenticateTries[ hostIndex ] = 0;
    localFloorStatus[ hostIndex] = initialFloorStatus;
	  reconnected[ hostIndex ] = false;
    nameExistsAgain[ hostIndex ] = false;
    unzipIsReady[ hostIndex ] = false;
    protocolIsTCP[ hostIndex ] = false;
    thisHostUpload[ hostIndex ] = null;
    thisHostUploadThread[ hostIndex ] = null;

  } // end initializeHostTableAt()

  // utility method to add an address to an empty position the host table
  // returns position index if succesful; -1 otherwise (no space in table)
  // assumes the address is not in the host table (does not verify)
  // the host table's use is synchronized by two arrays:
  //
  // hostInetAddr[hostnum] is null when the position is not in use, and holds
  //                the Internet address of a host when allocated
  // tunnelClosing[hostnum] is false when the tunnel is available
  //                for use; true when locked for closing
  public int addToHostTable(InetAddress newAddress,
                            String newSessionVariable,
                            int otherEndSetupPort,
                            int maybeNATPort)
  {
    // hack to convert loopback to local host address
    // permits running client & server on the same host
    String hostAddress = newAddress.getHostAddress();
    if(hostAddress.equals(loopbackHostAddress))
    {
      newAddress = localAddress;
      hostAddress = localHostAddress;
    }
    // look for a null position in the table
    int newIndex = videoAddressConfigured ? 1 : 0;
    synchronized(hostInetAddr)
    {
      for (; newIndex < maxConnections; ++newIndex)
        if (!tunnelClosing[newIndex])
          if (hostInetAddr[newIndex] == null)
            break;
      if (newIndex >= maxConnections)
        return -1;

      // insert the address in the table
      hostInetAddr[newIndex] = newAddress;
      hostAddr[newIndex] = hostAddress;
      sessionVariable[newIndex] = newSessionVariable;
      otherEndPort[newIndex] = otherEndSetupPort;
      sourcePort[newIndex] = maybeNATPort;
    }
    return newIndex;
    
  } // end addToHostTable()
  
  
  // utility method to find a username's location in host table
  // returns the index in table or -1 if not found
  public int findUsernameInHostTable(String nameSought)
  {
    String matchUsername = padto8(nameSought);
	  for(int hostIndex = 0; hostIndex < maxConnections; ++hostIndex)
	    if(username[hostIndex].equals(matchUsername))
	      return hostIndex;
	  return -1;
  }// end findUsernameInHostTable()
  
  
  // checks for inactive host table entry matching this user and clears it
  // (this can happen because we hold disconnected positions to enable 
  // auto-reconnect)
  void checkForDeadHostTableEntry(String offeredUsername)
  {
    // check for disconnected user of that name; if found, clean it out of table
	  int checkIndex = findUsernameInHostTable(offeredUsername);
    if(checkIndex > -1)
    {
      if(!tunnelIsActive[checkIndex])
      {
        // clean out the table position to allow re-use
        clearUserDataFromHostTable(checkIndex);
        clearHostDataFromHostTable(checkIndex); 
      }
    }
  }// end checkForDeadHostTableEntry()
  
  
  // clear out host table data that is not cleared by closeTunnel
  public static void clearUserDataFromHostTable(int hostIndex)
  {
	  username[hostIndex] = "        ";
	  if (hostIndex > highestReservedHost) 
	  {
      controlPrivilege[hostIndex] = '0';
      instructorPrivilege[hostIndex] = 0;
      drawGrid[hostIndex] = false;
    }
    appFileMtime[ hostIndex ] = 0;
    appFileSize[ hostIndex ] = 0;
    appFileStatNew[ hostIndex ] = false;
    netCapacitykbps[ hostIndex ] = 0;
        
    // if this host had floor, reset floorHolder
    int checkGroup = breakoutGroup[hostIndex];
    if(checkGroup >= 0)
    {
      if(breakoutFloorHolder[checkGroup] == hostIndex)
        breakoutFloorHolder[checkGroup] = -1;
    }
    if(floorHolder == hostIndex)
      floorHolder = -1;
    breakoutGroup[hostIndex] = 0;

  }// end clearUserDataFromHostTable()
  
  
  // clear host address data from host table
  public static void clearHostDataFromHostTable(int hostIndex)
  {
    if(tcpHandoffs[hostIndex]  != null)
    {
      if(tcpHandoffs[hostIndex].runningWait)
        tcpHandoffs[hostIndex].runAfterWait.interrupt();
      tcpHandoffs[hostIndex] = null;
    }
	  synchronized(hostInetAddr)
	  {
	    hostInetAddr[hostIndex] = null;
	    hostAddr[hostIndex] = null;
	  }
  }// end clearHostDataFromHostTable()
  
  
  // change current count of users and controllers
  // NOTE: static method serializes invocations
  public static void setUserCounts(int tunnels,int controllers)
  {
    numberOfTunnelsActive = tunnels;
    numberOfControllersLoggedIn = controllers;
      
    // update database
    updateDBNumUsers(numberOfTunnelsActive,numberOfControllersLoggedIn);
      
  }
  public static void incrementUserCounts(int hostIndex)
  {
    numberOfTunnelsActive++;
    if(controlPrivilege[hostIndex] != '0')
      numberOfControllersLoggedIn++;
      
    // update database
    updateDBNumUsers(numberOfTunnelsActive,numberOfControllersLoggedIn);
      
  }
  public static void decrementUserCounts(int hostIndex)
  {
    numberOfTunnelsActive--;
    if(controlPrivilege[hostIndex] != '0')
      numberOfControllersLoggedIn--;
      
    // update database
    updateDBNumUsers(numberOfTunnelsActive,numberOfControllersLoggedIn);
      
  } 
  
  /********************* format conversion utilities *************************/

  // turn contents of two bytes into an unsigned short
  public int bytesToUshort( byte byte1, byte byte2 )
  {
    int highByte = byte1;
    if(highByte < 0)highByte = 127 - highByte;
    int lowByte = byte2;
    if(lowByte < 0)lowByte = 127 - lowByte;
    return 256*highByte + lowByte;
  }
  public int bytesToUshort( byte[] bytes)
  {
    return bytesToUshort( bytes[0],bytes[1]);
  } // end bytesToUshort method
  

  // create the first byte of two from a short int
  static public byte intToByte1( int input )
  {
    byte byte1;
    int highByte = input >> 8;
    if( highByte <= 127 )byte1 = (byte) highByte;
    else byte1 = (byte)(int)(127-highByte);
    return byte1;
  } // end intToByte1 method


  // create the second byte of two from a short int
  static public byte intToByte2( int input )
  {
    byte byte1, byte2;
    int highByte = input >> 8;
    if( highByte <= 127 )byte1 = (byte) highByte;
    else byte1 = (byte)(int)(127-highByte);
    int lowByte = input - 256*highByte;
    if( lowByte <= 127 )byte2 = (byte)lowByte;
    else byte2 = (byte)(int)(127-lowByte);
    return byte2;
  } // end intToByte2 method


  // utility method to convert four bytes to a long
  public int bytesToUint(byte[] bytes, int offset)
  {
    int int1 =
      ( bytes[offset] >= 0 ? bytes[offset] : 255 - ~bytes[offset] ) * 16777216 +
      ( bytes[offset+1] >= 0 ? bytes[offset+1] : 255 - ~bytes[offset+1] ) * 65536 +
      ( bytes[offset+2] >= 0 ? bytes[offset+2] : 255 - ~bytes[offset+2] )  * 256 +
      ( bytes[offset+3] >= 0 ? bytes[offset+3] : 255 - ~bytes[offset+3] );
    return int1;
  } // end bytesToUint


  // encodes the specified value, returning the result as a 4 byte array.
  static byte[] intToBytes(int value)
  {
    ByteArrayOutputStream baos = new ByteArrayOutputStream(4);
    DataOutputStream dos = new DataOutputStream(baos);

    try
    {
      dos.writeInt(value);
    }
      catch(IOException ioe)
    {}

    return baos.toByteArray();
    
  } // end intToBytes()
  
  
  // encodes the specified value, returning the result as 8 byte array.
  static byte[] longToBytes(long value)
  {
    ByteArrayOutputStream baos = new ByteArrayOutputStream(4);
    DataOutputStream dos = new DataOutputStream(baos);

    try
    {
      dos.writeLong(value);
    }
      catch(IOException ioe)
    {}

    return baos.toByteArray();
    
  } // end longToBytes()
  
    
  // decodes the specified value, returning result of 8 bytes as long int 
  public long bytesToLong(byte[] buffer,int offset)
  {
    ByteArrayInputStream bais = new ByteArrayInputStream(buffer,offset,8);
    DataInputStream dis = new DataInputStream(bais);
    
    try
    {
      return dis.readLong();
    }
    catch(IOException ioe)
    {
      System.out.println("could not decode long");
      return 0;
    }
  }// end bytesToLong
  
  
  // method to pad out a String to 8 chars (used with username)
  // if longer than 8 chars the sring is truncated
  String padto8(String input)
  {
    String result = new String(input);
    while(result.length() < 8)result += ' ';
	return result.substring(0,8);
  }// end padto8()
  
  
  // method to pad out a String to 40 chars (used with fullname)
  // if longer than 8 chars the sring is truncated
  String padto40(String input)
  {
    String result;
    if(input == null)
	    result = " ";
	  else
	    result = new String(input);
    while(result.length() < 40)result += ' ';
	  return result.substring(0,40);
    
  }// end padto40()


  // method to return the date and time
  static String dateTime()
  {
    GregorianCalendar gcalendar = new GregorianCalendar();
    int year = gcalendar.get(Calendar.YEAR);
    int month = gcalendar.get(Calendar.MONTH) + 1 ;
    int date = gcalendar.get(Calendar.DATE);
    int hour = gcalendar.get(Calendar.HOUR_OF_DAY);
    int minute = gcalendar.get(Calendar.MINUTE);
    int second = gcalendar.get(Calendar.SECOND);
    String strMonth = "" + ( month < 10 ? "0" + month : "" + month);
    String strDate = "" + ( date < 10 ? "0" + date : "" + date);
    String strHour = "" + ( hour < 10 ? "0" + hour : "" + hour );
    String strMinute = "" + ( minute < 10 ? "0" + minute : "" + minute );
    String strSecond = "" + ( second < 10 ? "0" + second : "" + second );
    return year + "." + strMonth + "." + strDate + "/" + strHour + ":"
      + strMinute + ":" + strSecond;
  } // end dateTime method
  
  /********************** database - connection utilities **********************/

  // method to insert a log entry in NEWADMIN database
  synchronized static void addToDatabase(
                     String transactionType, long elapsedConnectTime,
                     String thisUsername, String thisIPAddress,
                     String thisUserCourseID, char controllerStatus,
                     String protocolCode, int peakRTT)
  {
    // data objects for MySQL
    if(noDatabase)return;
    Connection databaseConnection = null;
    DBQuery dbq = null;

    // connect to database
    try
    {
//      BufferedReader databaseInfoFile =
//      new BufferedReader( new FileReader( "NEWLIVE.dbkey" ));
//      String databaseAccount = databaseInfoFile.readLine();
//      databaseInfoFile.close();
//      int slashLocation = databaseAccount.indexOf("/");
//      String databasePassword = databaseAccount.substring(slashLocation+1);
//      databaseAccount = databaseAccount.substring(0,slashLocation);
      String databaseURL =
        "jdbc:mysql://" + databaseHost + "/" + databaseName + "?user=" +
         databaseUsername;
      dbq = new DBQuery("jdbc:mysql://" + databaseHost +"/" + databaseName, 
                        databaseUsername,databasePassword,databasePrefix);
    }
    catch (java.lang.ClassNotFoundException ee)
    {
      warnNoDatabase("MySQL access ClassNotFound" + ee);
    }
    catch(SQLException ex)
    {
      warnNoDatabase("MySQL exception"+ ex);
    }



    // create query
    try
    {
        Calendar rightNow = Calendar.getInstance();
        Long x = (rightNow.getTimeInMillis())/1000;
      
      dbq.insertLog(roomID, thisUsername.trim(), x.toString(), thisIPAddress, thisSemester,
                    thisUserCourseID.trim(), " ", String.valueOf(elapsedConnectTime),
                    String.valueOf(controllerStatus), String.valueOf(protocolCode),
                    String.valueOf(peakRTT), String.valueOf(numberOfTunnelsActive), transactionType);
    }
    catch(SQLException ex)
    {
      System.out.println("MySQL exception while posting statistics to database:"
        + ex.getMessage());
    } // end catch

    // close database connection
    dbq.close();
    
  } // end addToDatabase

  //nick: Updating the number of users tracked in the database
  static synchronized void updateDBNumUsers(int userCount, int controllerCount) {
    try 
    {
	    DBQuery dbq;
      String databaseURL =
          "jdbc:mysql://" + databaseHost + "/" + databaseName + "?user=" +
          databaseUsername;

      dbq = new DBQuery("jdbc:mysql://" + databaseHost + "/" + databaseName,
                        databaseUsername, databasePassword, databasePrefix); //nick
      dbq.postNumUsers(roomID,userCount);
      dbq.postNumPrivUsers(roomID,controllerCount);
	  dbq.close();
      return;				
    }
    catch (java.lang.ClassNotFoundException ee) 
    {
      warnNoDatabase("MySQL access ClassNotFound:" + ee);
	  return;
    }
    catch (SQLException ex) 
    {
      warnNoDatabase("MySQL exception:" + ex);
	  return;
    }
	
  }

  // method to warn database not found
  static synchronized void warnNoDatabase(String errorMessage)
  {
    if(noDatabase)return;
    System.out.println(
      "NEW TLM unable to open MySQL database:" + errorMessage);
    noDatabase = true;
  } // end warnNoDatabase()


  // method to look up entry in host table that has not yet authenticated
  //
  // returns the host table position in UAddr[] where the
  // address if found, or -1 if no matching address found
  static int lookupPendingConnection(InetAddress lookupAddress,
                                     String lookupSessionVariable,
                                     int lookupPort)
  {
    int hostNumber = 0;
    String hostAddress = lookupAddress.getHostAddress();

    // if there is a video source ignore that entry
    if( videoAddressConfigured )hostNumber = 1;

    // hostAddr is needed by other threads; make this quick!
    synchronized(hostAddr)
    {
      for(; hostNumber < maxConnections; hostNumber++)
      {
        if (hostAddr[hostNumber] != null)
          if (hostAddress.equals(hostAddr[hostNumber]))
            if (lookupPort == otherEndPort[hostNumber])
              if (lookupSessionVariable.equals(sessionVariable[hostNumber]) ||
                  lookupSessionVariable.equals("0000000000000000"))
                return hostNumber;
      }
    }// end synchronized(hostAddr)
    return -1;
  }
  

  // method to determine if a username is part of an active connection
  boolean userIsConnected(String testUsername)
  {
    for(int scan=0; scan < numberOfHosts; ++scan)
      if(hostInetAddr[scan] != null)
	    if(tunnelIsActive[scan])
          if(username[scan] != null)
            if(username[scan].equals(testUsername))
              return true;
    return false;
  }


  // method to decode and display error message from client
  public void receivedErrorMessage( byte[] recvBuffer, int remoteHost )
  {
    if(recvBuffer[3] != lastErrorMessageType[remoteHost])
      System.out.println( "Error message from " + hostAddr[remoteHost]
        + " TLM error code:" + recvBuffer[3]
        + " parameter:" + bytesToUshort( recvBuffer[4], recvBuffer[5] ));
    lastErrorMessageType[remoteHost] = recvBuffer[3];
  } // end receivedErrorMessage method
  
  
  // utility to convert a string to a valid filename
  public static String makeFileName(String moodleShortName)
  {
    String newFileName = moodleShortName.replace(' ','_');
	  newFileName = newFileName.replace("/",".");
	  newFileName = newFileName.replace("\\",".");
	
	  // insert here any other characters to be suppressed
    
    // truncate at last "-" in positions 6 to 11
    int lastDash = newFileName.lastIndexOf('-');
    if(lastDash > 6)newFileName = newFileName.substring(0,lastDash);
    
    return newFileName;
  }
  
  
/************************* sendRecordingsList **************************/
  
  // method to create and transmit list of server-side recordings 
  // when requested by a client
  void sendRecordingsList(int thisHostIndex, int fcMessageLength)
  {
    // only controllers are allowed to do this; also, can't
    // do if there are no recordings or download is in progress
    if((recordingsSubdirectory == null) || 
       (controlPrivilege[thisHostIndex] == '0') ||
       serverRecordingDownload)
    {
      fcConnect.sendToFC(50,0,null,thisHostIndex);
      return;
    }
    
    // there are two form of response: older client wants only filenames
    // and the fcMessageLength it sense = 0; effective with 5.2.4 the
    // client also wants filesize and it sends fcMessageLength = 1.
    int outputOffset = (fcMessageLength == 0) ? 0 : 8;
    
    // build a concatenated string of file lengths and names, separated by /
    File recordingsPath = new File(recordingsSubdirectory);
    if(! recordingsPath.exists())
    {
      fcConnect.sendToFC(50,0,null,thisHostIndex);
      return;
    }    
    File[] checkFiles = recordingsPath.listFiles();
    if(checkFiles.length == 0)
    {
      fcConnect.sendToFC(50,0,null,thisHostIndex);
      return;
    }
    String recordingsList = "", stringFilesize = "";
    for (int fileNumber = 0;
         fileNumber < checkFiles.length;
         ++fileNumber)
    {
      if(outputOffset > 0)
      {
        // make an 8 character string from filesize, with leading blanks to display
        int nextRecordingSize = (int)checkFiles[fileNumber].length();
        stringFilesize = new Integer(nextRecordingSize).toString();
        if(stringFilesize.length() > 8)
          stringFilesize = stringFilesize.substring(0,8);
        while(stringFilesize.length() < 8)stringFilesize = ' ' + stringFilesize;
      }
    
      String nextRecordingFilename = checkFiles[fileNumber].getName();
      if(nextRecordingFilename.endsWith("nwr"))
        // don't send list longer than UDP MTU over Ethernet
        if(recordingsList.length() + nextRecordingFilename.length() + outputOffset < 1454)
        {
          if(!nextRecordingFilename.equals(serverRecordingFileName))
            if(recordingsList.length() == 0)
            {
              if(outputOffset == 0)
                recordingsList = nextRecordingFilename;
              else
                recordingsList = stringFilesize + nextRecordingFilename;
            }
            else
            {
              if(outputOffset == 0)
                recordingsList = recordingsList + "/" + nextRecordingFilename;
              else
                recordingsList = recordingsList + "/" + stringFilesize + nextRecordingFilename;
            }
        }
        else
        {
          System.out.println("*** serverRecordings download truncated before " + 
            nextRecordingFilename + " - consider cleaning up the list!");
          break;
        }
    }
    
    // send the list to the requesting FC
    if(recordingsList.length() == 0)
    {
      fcConnect.sendToFC(50,0,null,thisHostIndex);
      return;
    }
    byte[] recordingsListBuffer = recordingsList.getBytes();
    fcConnect.sendToFC(50,
                       recordingsListBuffer.length,
                       recordingsListBuffer,
                       thisHostIndex);
    
  }// end sendRecordingsList()
  
} // end TLM class

