/* 
 * UploadRecording.java
 * Transport Layer Multicaster
 * Created on 1/14/13
 * @author  mpullen & students
 * @version 5.2.4
 */

/*----------------------------------------------------------------*
|  Copyright 2011-2013 Networking and Simulation Laboratory       |
|         George Mason University, Fairfax, Virginia              |
|                                                                 |
| Permission to use, copy, modify, and distribute this            |
| software and its documentation for academic purposes is hereby  |
| granted without fee, provided that the above copyright notice   |
| and this permission appear in all copies and in supporting      |
| documentation, and that the name of George Mason University     |
| not be used in advertising or publicity pertaining to           |
| distribution of the software without specific, written prior    |
| permission. GMU makes no representations about the suitability  |
| of this software for any purposes.  It is provided "AS IS"      |
| without express or implied warranties.  All risk associated     |
| with use of this software is expressly assumed by the user.     |
*----------------------------------------------------------------*/
//package TLM;
import java.sql.*;
import java.io.*;
import java.net.*;
import java.util.*;
import java.nio.channels.SocketChannel;
import java.nio.ByteBuffer;
import java.nio.channels.ClosedChannelException;
import java.nio.channels.ClosedByInterruptException;

// thread to make an uploaded recording file
public class UploadRecording extends TLM
{
  String filePathname;
  int thisHostIndex;
  int ackNo = 0;
  byte[] ackMessage = {66,0,5,0,0,121,0,2,0,0};
  FileOutputStream outFile;
  boolean terminateUpload = false;
  RecordQueueElement recordQueue = null;
  Thread recordingThread;
  MakeTheRecording makeTheRecording;
  SocketChannel thisSocketChannel;
    
  public UploadRecording(String filename,int invokingHostIndex)
  {
    filePathname = recordingsDownloadDirectory + '/' + filename;
    thisHostIndex = invokingHostIndex;
    thisSocketChannel = tcpSocketChannels[thisHostIndex];
    makeTheRecording = new MakeTheRecording(); 
    recordingThread = new Thread(makeTheRecording);
    recordingThread.start();
  }
  
  
  // internal class to make the recording
  class MakeTheRecording extends Thread
  {
    // run() method build the file and adds to it as messages arrive
    public void run()
    {
      // print request to log
      System.out.println(
        "recording upload requested by host:" + thisHostIndex +
        "\ncreating file:" + filePathname);                
      
      // make sure the directory configured is there
      File copyToDirectory = new File(recordingsDownloadDirectory);
      if(!copyToDirectory.exists())
      {
        System.err.println(
          "making directory to receive copy:" +
          recordingsDownloadDirectory);
        copyToDirectory.mkdir();
      }
      if(!copyToDirectory.isDirectory())
      {
        System.err.println(
          "*** configured copy to directory not a directory:" +
          recordingsDownloadDirectory);
        return;
      }
      
      // if upload file exists, delete it
      // unless it's a directory
      File checkFile = new File(filePathname);
      if(checkFile.exists())
      {
        if(checkFile.isDirectory())
        {
          System.err.println(
            "*** found directory of same name as file - could not upload");
          return;
        }
        System.out.println("deleting existing file of same name");
        checkFile.delete();
      }
      
      // create the recording file
      try
      {
        outFile = new FileOutputStream(filePathname);
      }
      catch(IOException ioe)
      {
        System.err.println(
          "*** unable to create requested upload file:" +
          filePathname);
        return;
      }
      
      // respond to requester accepting the upload
      fcConnect.sendToFC(120,0,null,thisHostIndex);
      
      // accept file segments as they become available 
      // write to file and ack them
      while(!stopTLM)
      {
        if(recordQueue != null)
        {
          // check for end of data code
          if(terminateUpload)return;
          if(recordQueue.queueMessage == null)
            setTerminate();
          else if(recordQueue.queueMessage.length == 0)
            setTerminate();
          else
          {
            synchronized(makeTheRecording)
            {
              try
              {
                outFile.write(recordQueue.queueMessage);
                ackMessage[8] = intToByte1(ackNo);
                ackMessage[9] = intToByte2(ackNo++);
                writeThisOutStream(ackMessage,10);
              }
              catch(IOException ioe)
              {
                System.err.println(
                  "*** IOException writing upload file:" +
                  filePathname);
                return;
              }
              recordQueue = recordQueue.nextInQueue;
            
            }// end synchronized(makeTheRecording)
          }// end else/if(recordQueue.queueMessage.length == 0)
          
          // check for end of data
          if(terminateUpload)
          {
            try
            {
              outFile.close();
            }
            catch(IOException ioe)
            {
              System.err.println(
                "*** IOException:" + ioe +
                "\nclosing file:" + filePathname);
              return;
            }
            System.out.println(
              "completed upload by host:" + thisHostIndex +
              " of file:" + filePathname + 
              " length:" + checkFile.length() + " bytes" );
            thisHostUpload[thisHostIndex] = null;
            return;  
          
          }// end if(terminateUpload)
          
        }// end if(recordQueue.queueMessage != null)
      
        // if nothing to write, wait until more is queued
        else 
        {
          try
          {
            synchronized(makeTheRecording)
            {
              wait();
            }
          }
          catch(InterruptedException ie)
          {
            // in case we need to stop
            terminateUpload = true;
          }
        }// end else
      
      }// end while(!stopTLM)
    
    }//end run()
    
  }// end class MakeTheRecording  
    
  void queueMessage(byte[] message)
  {
    synchronized(makeTheRecording)
    {
      // check for system stop condition
      if(stopTLM)return;
      
      // add message to the queue
      RecordQueueElement newQueueElement =
        new RecordQueueElement(message);
      if(recordQueue == null)
        recordQueue = newQueueElement;
      else
      {
        // find end of queue to add element
        RecordQueueElement chainToEnd = recordQueue;
        while(chainToEnd.nextInQueue != null)
          chainToEnd = chainToEnd.nextInQueue;
        chainToEnd.nextInQueue = newQueueElement;
      }
    }// end synchronized(makeTheRecording)
  }// end queueMessage()
  
  
  // notify the class to stop waiting
  void notifyWait()
  {
    synchronized(makeTheRecording)
    {
      makeTheRecording.notify();
    }
  }
  
  
  // set flag to terminate when queue is exhausted
  void setTerminate()
  {
    terminateUpload = true;
  }
  
  
  // writes the data to non-blocking channel
  // uses sleep to avoid lockup when receiver's buffer is full
  // returns true if disconnected or interrupted
  boolean writeThisOutStream(byte[] data,int length)
  {
    // move data to a ByteBuffer to send
    ByteBuffer bb = ByteBuffer.wrap(data,0,length);
  
    // keep trying until the buffer gets sent
    while(!stopTLM)
    {
      // write the data to outstream
	    try
	    {
        thisSocketChannel.write(bb);
      }
      catch(ClosedChannelException cce)
      {
        return true;
      }
      catch(Exception e)
      {
	      closeTunnel(thisHostIndex,false,
          "Exception sending ACK in UploadRecording for host:" + thisHostIndex + ":" + e);
        e.printStackTrace(System.out);  
	      return true;
      }
      
      // if data was not all written, wait sleepTime and try again
      if(bb.hasRemaining())
        try{Thread.sleep(sleepTime);}
        catch(InterruptedException ie)
        {
          return true;
        }
      else break;
    
    }// end while(!stopTLM)
       
    return false;
      
  }// end writeThisOutStream()
  
  
  // internal class to queue messages for recording
  class RecordQueueElement
  {
    byte[] queueMessage;
	  RecordQueueElement nextInQueue = null;
	
	  public RecordQueueElement(byte[] messageToQueue)
	  {
	    queueMessage = messageToQueue;
	  }
    
  }// end internal class RecordQueueElement
  
  
}// end class UploadRecording
